(*
Copyright 2009 Christian Sternagel, Ren Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory CetaVerifier imports ProofChecker Xml begin

abbreviation dropWhiteSpace :: "string \<Rightarrow> string" 
where "dropWhiteSpace \<equiv> filter (\<lambda> c. \<not> c mem wspace)"

definition eqWhiteSpace :: "string \<Rightarrow> string \<Rightarrow> bool" ("_ =w _" 50)
where "s1 =w s2 \<equiv> debug ''0''
  (''comparing whether parsed input corresponds to real input'')
  (dropWhiteSpace s1 = dropWhiteSpace s2)"

abbreviation s_goal :: string where "s_goal \<equiv> ''<goal>''"
abbreviation e_goal :: string where "e_goal \<equiv> ''</goal>''"
abbreviation s_input :: string where "s_input \<equiv> ''<input>''"
abbreviation e_input :: string where "e_input \<equiv> ''</input>''"

fun search_len :: "string \<Rightarrow> string \<Rightarrow> nat \<Rightarrow> (string \<times> string)option"
where "search_len t s n = (if take n t = s then Some ([],drop n t) 
                else (case t of Nil \<Rightarrow> None | Cons h tt \<Rightarrow> (case search_len tt s n of None \<Rightarrow> None | Some (bef,aft) \<Rightarrow> Some (h # bef,aft))))"

lemma search_len_sound:
  fixes t s before after::string
  assumes "search_len t s n = Some (before,after)"
  shows "t = (before @ s @ after)"
using assms
proof (induct t arbitrary: before)
  case Nil
  thus ?case 
    by (cases s, auto)
next
  case (Cons h tt)
  show ?case 
  proof (cases "take n (h # tt) = s")
    case True
    with Cons show ?thesis by auto 
  next
    case False with `search_len (h # tt) s n = Some (before,after)` 
    have res: "(case search_len tt s n of None \<Rightarrow> None | Some (bef,aft) \<Rightarrow> Some (h # bef, aft)) = Some (before,after)" by auto
    show ?thesis 
    proof (cases "search_len tt s n")
      case None with res show ?thesis by auto
    next
      case (Some befaft)
      show ?thesis proof (cases befaft)
	case (Pair bef aft)
	with Some res have "search_len tt s n = Some (bef,aft) \<and> before = h # bef \<and> after = aft" by auto
	with Cons show ?thesis by auto
      qed
    qed
  qed
qed

declare search_len.simps[simp del]    

abbreviation search :: "string \<Rightarrow> string \<Rightarrow> (string \<times> string)option"
where "search t s \<equiv> search_len t s (length s)"
 

fun extract_goal_input :: "string \<Rightarrow> string option"
where "extract_goal_input s = debug ''0'' ''extracting goal and TRS from input string'' (
   case search s s_goal of
      None \<Rightarrow> None
   | Some (_,mgoal) \<Rightarrow> (
   case search mgoal e_goal of 
      None \<Rightarrow> None
   | Some (goal,s') \<Rightarrow> (
   case search s' s_input of
      None \<Rightarrow> None
   | Some (_,minput) \<Rightarrow> (
   case search minput e_input of
      None \<Rightarrow> None
   | Some (input,_) \<Rightarrow> Some (concat [''<problem>'',s_goal,goal,e_goal,s_input,input,e_input,''</problem>''])
   ))))"


lemma extract_goal_input_sound: 
  fixes append :: "string \<Rightarrow> string \<Rightarrow> string" (infixr "<@>" 65)
  defines [simp]: "append \<equiv> List.append"
  assumes res: "extract_goal_input s = Some res"
  shows "\<exists> bef goal mid input aft. 
                 res = (''<problem><goal>'' <@> goal <@> ''</goal><input>'' <@> input <@> ''</input></problem>'' ) \<and> 
                   s = (bef <@> ''<goal>'' <@> goal <@> ''</goal>'' <@> mid <@> ''<input>'' <@> input <@> ''</input>'' <@> aft)"
proof (cases "search s s_goal")
  case None
  with res show ?thesis by auto
next
  case (Some bef_mgoal)
  show ?thesis
  proof (cases bef_mgoal)
    case (Pair bef mgoal)
    with Some have s1: "search s s_goal = Some (bef, mgoal)" by auto
    show ?thesis 
    proof (cases "search mgoal e_goal")
      case None
      with res s1 show ?thesis by auto
    next
      case (Some goal_sp)
      show ?thesis
      proof (cases goal_sp)
	case (Pair goal s')
	with Some s1 have s2: "search mgoal e_goal = Some (goal, s')" by auto
	show ?thesis
	proof (cases "search s' s_input")
	  case None 
	  with res s1 s2 show ?thesis by auto
	next
	  case (Some mid_minput)
	  show ?thesis 
	  proof (cases mid_minput)
	    case (Pair mid minput)
	    with Some s1 s2 have s3: "search s' s_input = Some (mid, minput)" by auto
	    show ?thesis 
	    proof (cases "search minput e_input")
	      case None
	      with res s1 s2 s3 show ?thesis by auto
	    next
	      case (Some input_aft)
	      show ?thesis
	      proof (cases input_aft)
		case (Pair input aft)
		with Some s1 s2 s3 have s4: "search minput e_input = Some (input,aft)" by auto
		
		from s1 have sp1: "s = (bef @ s_goal @ mgoal)" by (rule search_len_sound)
		from s2 have sp2: "mgoal = (goal @ e_goal @ s')" by (rule search_len_sound)
		from s3 have sp3: "s' = (mid @ s_input @ minput)" by (rule search_len_sound)
		from s4 have sp4: "minput = (input @ e_input @ aft)" by (rule search_len_sound)

		from s1 s2 s3 s4 res 
		have correctRes: "res = (''<problem><goal>'' <@> goal <@> ''</goal><input>'' <@> input <@> ''</input></problem>'' )" by auto
		from sp1 sp2 sp3 sp4 
		have correctSplit: "s = (bef <@> ''<goal>'' <@> goal <@> ''</goal>'' <@> mid <@> ''<input>'' <@> input <@> ''</input>'' <@> aft)" by auto
		
		from correctRes correctSplit show ?thesis by blast
	      qed
	    qed
	  qed
	qed
      qed
    qed
  qed
qed

  
fun certifyProof :: "string \<Rightarrow> certificationResult"
where "certifyProof problemString  = (
  case parseCertProblem problemString of
     Left err \<Rightarrow> Error (''error while parsing'' @ [newline] @ err)
   | Right cp \<Rightarrow> (
  case extract_goal_input problemString of
    None \<Rightarrow> Error ''could not find goal and input in string for comparison with parsed result''
  | Some inpProblem \<Rightarrow> (
  if (inpProblem =w (show (xml cp))) then  certifyCertProblem cp else 
      Error (concat [''parsed problem does not correspond to input'', [newline], ''input:'',[newline], dropWhiteSpace inpProblem,[newline],[newline],''parsed:'',[newline],dropWhiteSpace (show (xml cp)), [newline]])
  )))"

text {*
if the certifier gives an answer then
- the input problem string contains a goal and an input
- moreover, there is a certification problem with some goal and some trs where the pretty printer delivers exactly the goal and the input of the problem string
- and for this trs the goal has been proven: the trs is terminating if the goal was True (termination), or it is non-terminating if the goal was False (non-termination)
- the answer is the same as the goal in the certification problem
*}
lemma certifyProofSound : 
  fixes append :: "string \<Rightarrow> string \<Rightarrow> string" (infixr "<@>" 65)
  defines [simp]: "append \<equiv> List.append"
  assumes answer: "certifyProof problemString = Certified answer"
  shows "\<exists> goal (trs :: (string,string)trsL) prof goals input bef mid aft. 
               problemString = (bef <@> ''<goal>'' <@> goals <@> ''</goal>'' <@> mid <@> ''<input>'' <@> input <@> ''</input>'' <@> aft) \<and>
               (''<problem><goal>'' <@> goals <@> ''</goal><input>'' <@> input <@> ''</input></problem>'') =w show (xml ((CertificationProblem goal trs prof) :: (string,labelType,string)certificationProblem)) \<and>
               SN (rstep (set trs)) = goal \<and> 
               answer = goal"
proof (cases "parseCertProblem problemString")
  case (Left err)
  with answer show ?thesis by auto
next
  case (Right cp)
  show ?thesis 
  proof (cases cp)
    case (CertificationProblem goal trs prof)
    with Right have cp: "parseCertProblem problemString = Right ((CertificationProblem goal trs prof) :: (string,labelType,string)certificationProblem)" by auto
    show ?thesis 
    proof (cases "extract_goal_input problemString")
      case None
      with cp answer show ?thesis by auto
    next
      case (Some inpProblem)
      show ?thesis
      proof (cases "inpProblem =w show (xml cp)")
	case False
	with answer Right Some show ?thesis by auto
      next
	case True
	with answer Right Some have "certifyCertProblem cp = Certified answer" by auto
	with certifyCertProblemSound[of goal trs prof answer]  CertificationProblem have correctAnswer: "SN (rstep (set trs)) = goal \<and> answer = goal" (is "?SN \<and> ?ansGoal") by auto
	from Some   have "\<exists> bef goals mid input aft. 
               inpProblem = (''<problem><goal>'' <@> goals <@> ''</goal><input>'' <@> input <@> ''</input></problem>'') \<and>
               problemString = (bef <@> ''<goal>'' <@> goals <@> ''</goal>'' <@> mid <@> ''<input>'' <@> input <@> ''</input>'' <@> aft)" 
	  unfolding append_def by (rule extract_goal_input_sound)
	from this obtain bef goals mid input aft where 
              "inpProblem = (''<problem><goal>'' <@> goals <@> ''</goal><input>'' <@> input <@> ''</input></problem>'') \<and>
               problemString = (bef <@> ''<goal>'' <@> goals <@> ''</goal>'' <@> mid <@> ''<input>'' <@> input <@> ''</input>'' <@> aft)" (is "?inp \<and> ?prob") by blast 
	hence ?inp and ?prob by auto
	from `?inp` CertificationProblem True have
           "(''<problem><goal>'' <@> goals <@> ''</goal><input>'' <@> input <@> ''</input></problem>'') =w show (xml (CertificationProblem goal trs prof))" (is ?eqwhite) by auto
	show ?thesis unfolding append_def
          by (intro exI,intro conjI,rule `?prob`[unfolded append_def],rule `?eqwhite`[unfolded append_def],auto simp: correctAnswer)
      qed
    qed
  qed
qed

end
