(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory RedPair imports Trs DpFramework begin


definition one_step_ctxt_closed :: "('f,'v)trs \<Rightarrow> bool"
where
  "one_step_ctxt_closed r = (\<forall> f bef s t aft. ((s,t) \<in> r \<longrightarrow> (Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> r))"

lemma one_step_ctxt_closed_imp_ctxt_closed: "one_step_ctxt_closed r \<Longrightarrow> ctxt_closed r"
proof (unfold one_step_ctxt_closed_def ctxt_closed_def)
  assume one: "(\<forall> f bef s t aft. ((s,t) \<in> r \<longrightarrow> (Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> r))"
  show "CC r \<subseteq> r"
  proof
    fix csct
    assume csct: "csct \<in> CC r"
    thus "csct \<in> r"
    proof (cases csct)
      case (Pair cs ct)
      with csct have "(cs, ct) \<in> CC r" by simp
      thus ?thesis
      proof
	fix s t C 
	assume cs: "cs = C\<langle>s\<rangle>" and ct: "ct = C\<langle>t\<rangle>" and st: "(s,t) \<in> r"
	from st have "(C\<langle>s\<rangle>,C\<langle>t\<rangle>) \<in> r" 
	proof (induct C arbitrary: s t)
	  case Hole
	  thus ?case by simp
	next
	  case (More f bef C aft)
	  with one show ?case by auto
	qed
	with Pair cs ct show ?thesis by simp 
      qed
    qed
  qed
qed

lemma ctxt_closed_imp_one_step_ctxt_closed : fixes r :: "('f,'v)trs"
  assumes "ctxt_closed r"
  shows "one_step_ctxt_closed r"
proof (unfold one_step_ctxt_closed_def, clarify)
    fix f :: 'f and  bef and s t :: "('f,'v)term" and aft :: "('f,'v) term list"
    assume step: "(s,t) \<in> r"
    let ?C = "More f bef \<box> aft"
    have "Fun f (bef @ s # aft) = ?C\<langle>s\<rangle>" (is "?l = _")
    and  "Fun f (bef @ t # aft) = ?C\<langle>t\<rangle>" (is "?r = _") by auto
    hence "(?l,?r) \<in> CC r" using step ccI[where s = s and t = t and r = r and C = ?C] by simp
    thus "(?l,?r) \<in> r" using assms unfolding ctxt_closed_def by auto
qed

lemma one_step_ctxt_closed_rstep: "one_step_ctxt_closed (rstep r)"
using ctxt_closed_imp_one_step_ctxt_closed ctxt_closed_rstep by auto


definition all_ctxt_closed :: "('f,'v)trs \<Rightarrow> bool"
where "all_ctxt_closed rel \<equiv> \<forall> f ts ss. (length ts = length ss \<and> (\<forall> i. i < length ts \<longrightarrow> (ts ! i, ss ! i) \<in> rel) \<longrightarrow> (Fun f ts, Fun f ss) \<in> rel)"


lemma tr_one_imp_all_ctxt_closed: assumes tran: "trans r" 
  and refl: "refl r"
  and one: "one_step_ctxt_closed r" 
  shows "all_ctxt_closed r"
unfolding all_ctxt_closed_def
proof (clarify)
  fix f ts ss
  assume l: "length ts = length ss" and
         steps: "\<forall> i < length ts. (ts ! i, ss ! i) \<in> r"
  let ?n = "length ts"
  let ?us = "\<lambda> i. take i ss @ (take (?n - i) (List.drop i ts))"
  have "\<And> i. i \<le> ?n \<Longrightarrow> (Fun f ts, Fun f (?us i)) \<in> r"
  proof -
    fix i
    assume "i \<le> ?n"
    thus "(Fun f ts, Fun f (?us i)) \<in> r"
    proof (induct i)
      case 0 thus ?case using refl unfolding refl_on_def by auto
    next
      case (Suc i)
      hence i: "i < length ts" by auto
      from Suc have ind: "(Fun f ts, Fun f (?us i)) \<in> r" by auto
      from Suc steps have step: "(ts ! i, ss ! i) \<in> r" by auto
      let ?bef = "take i ss"
      let ?aft = "take (?n - Suc i) (List.drop (Suc i) ts)"
      from one step have cstep: "(Fun f (?bef @ ts ! i # ?aft), Fun f (?bef @ ss ! i # ?aft)) \<in> r" (is "(?l,?r) \<in> r") unfolding one_step_ctxt_closed_def by auto
      have lId: "Fun f (?us i) = ?l" using l i by (auto simp: drop_Suc_conv_tl)
      have rId: "Fun f (?us (Suc i)) = ?r" using l i by (auto simp: take_Suc_conv_app_nth)
      from cstep have "(Fun f (?us i), Fun f (?us (Suc i))) \<in> r" by (simp only: lId rId)
      with  ind show ?case using tran unfolding trans_def by best
    qed
  qed      
  hence "(Fun f ts, Fun f (?us ?n)) \<in> r" by force
  thus "(Fun f ts, Fun f ss) \<in> r" using l by  auto
qed


lemma rtrancl_one_step_ctxt_closed : assumes 
  "one_step_ctxt_closed r" 
  shows "one_step_ctxt_closed (r^*)"
unfolding one_step_ctxt_closed_def 
proof clarify
  fix f bef s t aft
  assume "(s,t) \<in> r^*"
  thus "(Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> r^*"
  proof (induct rule: rtrancl_induct, simp)
    fix t u
    assume a: "(t,u) \<in> r" and
      b: "(Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> r^*"
    from a assms have "(Fun f (bef @ t # aft), Fun f (bef @ u # aft)) \<in> r" unfolding one_step_ctxt_closed_def by auto
    with b show "(Fun f (bef @ s # aft), Fun f (bef @ u # aft)) \<in> r^*" by auto
  qed
qed



lemma rstep_rtrancl_all_ctxt_closed: 
  shows "all_ctxt_closed ((rstep r)^*)"
proof -
  have a: "one_step_ctxt_closed ((rstep r)^*)"
    by (rule rtrancl_one_step_ctxt_closed, auto simp: one_step_ctxt_closed_rstep)
  show ?thesis 
    by (rule tr_one_imp_all_ctxt_closed, auto simp: trans_rtrancl refl_rtrancl a)
qed


lemma trs_inclusion : assumes C: "one_step_ctxt_closed Rel" 
  and s: "sub_closed Rel"
  and orient: "R \<subseteq> Rel" 
  shows "rstep R \<subseteq> Rel"
proof -
  from C have D: "ctxt_closed Rel" by (rule one_step_ctxt_closed_imp_ctxt_closed)
  from orient have "CS R \<subseteq> CS Rel" by (rule CS_monotone)
  with s have "CS R \<subseteq> Rel" by (auto simp: sub_closed_def)
  hence "rstep R \<subseteq> CC Rel" by (rule CC_monotone)
  with D show ?thesis by (auto simp: ctxt_closed_def)
qed
      

locale redpair =    
  fixes  S :: "('f,'v)trs" ("\<succ>")
    and   NS :: "('f,'v)trs" ("\<succeq>")
  assumes SN: "SN \<succ>"
    and compat: "\<succeq> O \<succ> \<subseteq> \<succ>"
    and wmNS: "one_step_ctxt_closed \<succeq>"
    and subNS: "sub_closed \<succeq>"
    and subS: "sub_closed \<succ>"
begin

lemma mNS: "ctxt_closed \<succeq>" 
  using wmNS by (rule one_step_ctxt_closed_imp_ctxt_closed) 

lemma trCompat : "\<succeq>^* O \<succ> \<subseteq> \<succ>" using compat by (rule compat_tr_compat)


lemma trans_split_NS_S_union: (* SW *)
  assumes "r^* \<subseteq> (\<succeq> \<union> \<succ>)^*"
  shows "r^* \<subseteq> \<succ> O \<succ>^* O \<succeq>^* \<union> \<succeq>^*"
proof(rule subsetI2)
  fix x y
  assume "(x, y)  \<in> r^*"
  with assms have  "(x,y) \<in> (\<succeq> \<union> \<succ>)^*" by auto
  thus "(x,y) \<in> \<succ> O \<succ>^* O \<succeq>^* \<union> \<succeq>^*" using split_NS_S_union[where S = S and NS = NS] compat by auto
qed

  
lemma redPairSound :
  assumes rules: "R \<subseteq> \<succeq>"
  and nonStrict: "rP \<inter> P \<subseteq> \<succeq> \<union> \<succ>"
  and strict: "P - rP \<subseteq> \<succ>"
  and chain: "ichain (P,R) s t \<sigma>"
  shows  "\<exists> j. ichain (rP, R) (shift s j) (shift t j) (shift \<sigma> j)"
using chain
proof auto
  let ?ss = "\<lambda> i. s i \<cdot> \<sigma> i"
  let ?ts = "\<lambda> i. t i \<cdot> \<sigma> i"
  let ?next = "\<lambda> i. (s (Suc i), t (Suc i))"
  assume allP: "\<forall> i. (s i, t i) \<in> P"
  assume steps: "\<forall> i. (?ts i, ?ss (Suc i)) \<in> (rstep R)^*"
  have piece1: "\<forall> i. (?ts i, ?ts (Suc i)) \<in> \<succ> \<or> (?ts i, ?ts (Suc i)) \<in> \<succeq>^* \<and> (?next i) \<in> rP"
  proof
    fix i
    from subNS wmNS rules trs_inclusion have "rstep R \<subseteq> \<succeq>" by auto
    hence "(rstep R)^* \<subseteq> \<succeq>^*" by (rule rtrancl_mono)
    with steps have stepsNS: "(?ts i, ?ss (Suc i)) \<in> \<succeq>^*" by blast
    show "(?ts i, ?ts (Suc i)) \<in> \<succ> \<or> (?ts i, ?ts (Suc i)) \<in> \<succeq>^* \<and> (?next i) \<in> rP" 
    proof (cases "?next i \<in> \<succ> \<inter> (P \<union> rP)")
      case True
      hence "(?ss (Suc i), ?ts (Suc i)) \<in> \<succ>" using subS by (auto simp: sub_closed_def)
      with stepsNS trCompat show ?thesis by auto 
    next
      case False
      with strict nonStrict allP
      have one: "?next i \<in> \<succeq>" and two: "?next i \<in> rP" by auto
      from one have "(?ss (Suc i), ?ts (Suc i)) \<in> \<succeq>" using subNS by (auto simp: sub_closed_def)
      with stepsNS have "(?ts i, ?ts (Suc i)) \<in> \<succeq>^*" by auto
      with two show ?thesis by simp
    qed
  qed
  
  hence infSeq: "\<forall> i. (?ts i, ?ts (Suc i)) \<in> \<succeq>^* \<union> \<succ>" by auto
  from SN have "SN_elt \<succ> (?ts 0)" by auto
  from infSeq trCompat this have "\<exists> j. \<forall> i \<ge> j. (?ts i, ?ts (Suc i)) \<in> \<succeq>^* - \<succ>" by (rule non_strict_ending)
  
  from this obtain j where "\<forall> i \<ge> j. (?ts i, ?ts (Suc i)) \<in> \<succeq>^* - \<succ>" ..
  with piece1 have "\<forall> i \<ge> j. ?next i \<in> rP" by blast
  hence "\<forall> i. (s (i+(Suc j)), t (i + (Suc j))) \<in> rP" by auto
  thus "\<exists> j. \<forall> i. (s (i+j), t (i+j)) \<in> rP" ..
qed
end


abbreviation redPairProc :: "('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> bool"
 where "redPairProc S NS P R rP rR \<equiv> R \<subseteq> NS \<and> rP \<inter> P \<subseteq> NS \<union> S \<and> P - rP \<subseteq> S \<and> rR = R"

context redpair
begin
lemma redPairProc_ci : "chain_identifying_proc (redPairProc \<succ> \<succeq>)"
unfolding chain_identifying_proc_def
proof ((rule allI)+, rule impI)
  fix P R rP rR
  assume proc: "redPairProc \<succ> \<succeq> P R rP rR"
  show "rR \<subseteq>ci R \<and> (\<forall> s t \<sigma> . min_ichain (P, R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (rP, rR) (shift s i) (shift t i) (shift \<sigma> i)))"
  proof (rule conjI, simp add: proc ci_subset_refl, (rule allI)+, rule impI)
    fix s t \<sigma>
    assume min: "min_ichain (P,R) s t \<sigma>"
    hence "ichain (P,R) s t \<sigma>" by simp
    with proc show "\<exists> j. ichain (rP, rR) (shift s j) (shift t j) (shift \<sigma> j)" using redPairSound by auto
  qed
qed


lemma monoRedPairSound : (* SW *)
  assumes monotonic: "one_step_ctxt_closed \<succ>"
  and rNonStrict: "rR \<inter> R \<subseteq> \<succeq> \<union> \<succ>"
  and rStrict: "R - rR \<subseteq> \<succ>"
  and pNonStrict: "rP \<inter> P \<subseteq> \<succeq> \<union> \<succ>"
  and pStrict: "P - rP \<subseteq> \<succ>"
  and chain: "ichain (P,R) s t \<sigma>"
  shows  "\<exists> j. ichain (rP, rR) (shift s j) (shift t j) (shift \<sigma> j)"
using chain
proof auto
  let ?ss = "\<lambda> i. s i \<cdot> \<sigma> i"
  let ?ts = "\<lambda> i. t i \<cdot> \<sigma> i"
  let ?nextr = "\<lambda> i. (?ts i, ?ss (Suc i))"
  let ?nextp = "\<lambda> i. (s i, t i)"
  assume allP: "\<forall> i. (s i, t i) \<in> P"
  assume steps: "\<forall> i. (?ts i, ?ss (Suc i)) \<in> (rstep R)^*"
  from monotonic wmNS have monoNSS:"one_step_ctxt_closed (\<succeq> \<union> \<succ>)" unfolding one_step_ctxt_closed_def by auto
  from subS subNS have subNSS:"sub_closed (\<succeq> \<union> \<succ>)" unfolding sub_closed_def using cs_union_distr[where ?R="\<succeq>" and ?S=\<succ>] by auto
  let ?SS = "\<succ> O \<succ>^* O \<succeq>^*"
  from rStrict rNonStrict have "R \<subseteq> \<succeq> \<union> \<succ>" by auto
  with monoNSS subNSS trs_inclusion have "(rstep R) \<subseteq> (\<succeq> \<union> \<succ>)" by auto
  hence "(rstep R)^* \<subseteq> (\<succeq> \<union> \<succ>)^*" by (rule rtrancl_mono)
  hence rsteps_in_NSS_or_NS:"(rstep R)^* \<subseteq> ?SS \<union> \<succeq>^*" using trans_split_NS_S_union by auto
  have piece1: "\<forall> i. (?ss i, ?ss (Suc i)) \<in> ?SS \<or> (?ss i, ?ss (Suc i)) \<in> \<succeq>^* \<and> ((?nextr i) \<in> (rstep rR)^* \<and> (?nextp i) \<in> rP)"
  proof
    fix i
    show "(?ss i, ?ss (Suc i)) \<in> ?SS \<or> (?ss i, ?ss (Suc i)) \<in> \<succeq>^* \<and> (?nextr i) \<in> (rstep rR)^* \<and> (?nextp i) \<in> rP"
    proof (cases "?nextp i \<in> \<succ> \<inter> (P \<union> rP)")
      (* case 1: strict decrease with DP step *)
      case True
      hence "(?ss i, ?ts i) \<in> \<succ>" using subS by (auto simp: sub_closed_def)
      from steps rsteps_in_NSS_or_NS have "(?nextr i) \<in> ?SS \<union> \<succeq>^*" by auto
      with `(?ss i, ?ts i) \<in> \<succ>` have "(?ss i, ?ss (Suc i)) \<in> \<succ> O (?SS \<union> \<succeq>^*)" by auto
      hence "(?ss i, ?ss (Suc i)) \<in> ?SS" by simp
      thus ?thesis by auto
    next
      case False
      with pStrict pNonStrict allP
      have "?nextp i \<in> \<succeq>" and "?nextp i \<in> rP" by auto
      hence dp_step_in_NS: "(?ss i, ?ts i) \<in> \<succeq>" using subNS by (auto simp: sub_closed_def)
      show "(?ss i, ?ss (Suc i)) \<in> ?SS \<or> (?ss i, ?ss (Suc i)) \<in> \<succeq>^* \<and> (?nextr i) \<in> (rstep rR)^* \<and> (?nextp i) \<in> rP"
      proof (cases "(?nextr i) \<in> ((rstep rR)^* - ?SS)")
        (* case 2: strict decrease in R steps *)
        case False
        hence or:"(?nextr i) \<notin> ((rstep rR)^*) \<or> (?nextr i) \<in> ?SS" by auto
        have "(?nextr i) \<in> ?SS"
        proof (cases "(?nextr i) \<in> ?SS", simp)
          case False
          with or steps have "(?nextr i) \<in> ((rstep R)^* ) - ((rstep rR)^* )" by auto
          hence "(?nextr i) \<in> (rstep R)^* O ((rstep R) - (rstep rR)) O (rstep R)^*"
            using rtrancl_diff_decomp[where ?A="rstep R"] by auto
          hence "\<exists> u v. (?ts i, u) \<in> (rstep R)^* \<and> (u, v) \<in> ((rstep R) - (rstep rR)) \<and> (v, ?ss (Suc i)) \<in> (rstep R)^*" by auto
          then obtain u v where
            tu:"(?ts i, u) \<in> (rstep R)^*" and uv:"(u, v) \<in> ((rstep R) - (rstep rR))" and vs:"(v, ?ss (Suc i)) \<in> (rstep R)^*" by auto
          with rsteps_in_NSS_or_NS have tu:"(?ts i, u) \<in> ?SS \<union> \<succeq>^*" and vs:"(v, ?ss (Suc i)) \<in> ?SS \<union> \<succeq>^*" by auto
          from rStrict monotonic subS trs_inclusion have "rstep (R - rR) \<subseteq> \<succ>" by auto
          with uv have "(u, v) \<in> \<succ>" using rstep_diff by auto
          with tu have "(?ts i, v) \<in> (?SS O \<succ>) \<union> (\<succeq>^* O \<succ>)" by auto
          with trCompat have "(?ts i, v) \<in> (\<succ> O  \<succ>^* O \<succeq>^* O \<succ>) \<union> \<succ>" by auto
          with trCompat have "(?ts i, v) \<in> (\<succ> O \<succ>^* O \<succ>) \<union> \<succ>" by auto
          hence "(?ts i, v) \<in> \<succ> O \<succ>^*" by auto
          with vs have nextrstep:"(?ts i, ?ss (Suc i)) \<in> (\<succ> O \<succ>^* O \<succ> O \<succ>^* O \<succeq>^*) \<union> (\<succ> O \<succ>^* O \<succeq>^*)" by auto
          have "\<succ> O \<succ>^* O \<succ> O \<succ>^* \<subseteq> \<succ> O (\<succ>^* O \<succ>^*)" using O_assoc[where ?R="\<succ>^*" and ?S="\<succ>^*" and ?T="\<succ>"] by auto
          hence "\<succ> O \<succ>^* O \<succ> O \<succ>^* \<subseteq> \<succ> O \<succ>^*" using rtrancl_idemp_self_comp by auto
          hence "(\<succ> O \<succ>^* O \<succ> O \<succ>^*) O \<succeq>^* \<subseteq> (\<succ> O \<succ>^*) O \<succeq>^*" by (rule rel_comp_mono) simp
          hence "(\<succ> O \<succ>^* O \<succ> O \<succ>^* O \<succeq>^*) \<union> (\<succ> O \<succ>^* O \<succeq>^*) \<subseteq> (\<succ> O \<succ>^* O \<succeq>^*)" by (auto simp: O_assoc)
          from this nextrstep show  "(?ts i, ?ss (Suc i)) \<in> \<succ> O \<succ>^* O \<succeq>^*" 
          by (rule set_mp)
        qed
        with dp_step_in_NS trCompat show ?thesis by auto
      next
        (* case 3: only weak decrease with both DP and R steps *)
        case True
        hence "(?nextr i) \<in> ((rstep rR)^*)" by auto
        from steps rsteps_in_NSS_or_NS True have "(?nextr i) \<in> \<succeq>^*" by auto
        from dp_step_in_NS `(?nextr i) \<in> \<succeq>^*` have  "(?ss i, ?ss (Suc i)) \<in> \<succeq>^*" by auto
        with `?nextp i \<in> rP` `(?nextr i) \<in> ((rstep rR)^*)` show ?thesis by simp
      qed
    qed
  qed
  hence infSeq: "\<forall> i. (?ss i, ?ss (Suc i)) \<in>  \<succeq>^* \<union> (\<succ> O \<succ>^* O \<succeq>^*)" by auto
  from trCompat have compat2:"\<succeq>^* O (\<succ> O \<succ>^* O \<succeq>^*) \<subseteq> (\<succ> O \<succ>^* O \<succeq>^*)" by auto
  have "SN (\<succ> O \<succ>^* O \<succeq>^*)" using NS_comp_S_is_SN[where S = S and NS = NS] SN compat by auto
  hence "SN_elt (\<succ> O \<succ>^* O \<succeq>^*) (?ss 0)" by auto
  from infSeq compat2 this have "\<exists> j. \<forall> i \<ge> j. (?ss i, ?ss (Suc i)) \<in> \<succeq>^* - (\<succ> O \<succ>^* O \<succeq>^*)" using non_strict_ending by auto
  from this obtain j where "\<forall> i \<ge> j. (?ss i, ?ss (Suc i)) \<in> \<succeq>^* - (\<succ> O \<succ>^* O \<succeq>^*)" ..
  with piece1 have "\<forall> i \<ge> j. (?nextr i \<in> ((rstep rR)^*) \<and> (?nextp i) \<in> rP)" by blast
  hence "\<forall> i. ((?ts (i + j), ?ss (Suc i + j)) \<in> ((rstep rR)^*)  \<and> (s (i + j), t (i + j)) \<in> rP)" by auto
  thus "\<exists>j. (\<forall>i. (s (i + j), t (i + j)) \<in> rP) \<and> (\<forall>i. (t (i + j) \<cdot> \<sigma> (i + j), s (Suc (i + j)) \<cdot> \<sigma> (Suc (i + j))) \<in> (rstep rR)\<^sup>*)" by auto
qed
end


abbreviation monoRedPairProc :: "('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> bool"
 where "monoRedPairProc S NS P R rP rR \<equiv> one_step_ctxt_closed S \<and> rR \<subseteq> R \<and> rR \<inter> R \<subseteq> NS \<union> S \<and> R - rR \<subseteq> S \<and> rP \<inter> P \<subseteq> NS \<union> S \<and> P - rP \<subseteq> S"

context redpair
begin
lemma monoRedPairProc_ci : "chain_identifying_proc (monoRedPairProc \<succ> \<succeq>)"
unfolding chain_identifying_proc_def
proof ((rule allI)+, rule impI)
  fix P R rP rR
  assume proc: "monoRedPairProc \<succ> \<succeq> P R rP rR"
  show "rR \<subseteq>ci R \<and> (\<forall> s t \<sigma> . min_ichain (P, R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (rP, rR) (shift s i) (shift t i) (shift \<sigma> i)))"
  proof (rule conjI, simp add: proc ci_subsetI[of rR R], (rule allI)+, rule impI)
    fix s t \<sigma>
    assume min: "min_ichain (P,R) s t \<sigma>"
    hence "ichain (P,R) s t \<sigma>" by simp
    with proc show "\<exists> j. ichain (rP, rR) (shift s j) (shift t j) (shift \<sigma> j)" using monoRedPairSound by auto
  qed
qed

lemma ruleRemoval:
  assumes monotonic: "one_step_ctxt_closed \<succ>"
  and nonStrict: "R \<subseteq> \<succeq> \<union> \<succ>"
  and strict: "R - rR \<subseteq> \<succ>"
  and subset:"rR \<subseteq> R"
  and ideriv: "ideriv R t"
  shows "\<exists> i. ideriv rR (shift t i)"
proof -
  from ideriv have alli:"\<forall>i. (t i, t (Suc i)) \<in> rstep R" unfolding ideriv_def by auto
  from SN have "SN_elt \<succ> (t 0)" by auto
  from monotonic strict subS trs_inclusion have "(rstep (R - rR)) \<subseteq> \<succ>" by auto
  from monotonic wmNS have monoNSS:"one_step_ctxt_closed (\<succeq> \<union> \<succ>)" unfolding one_step_ctxt_closed_def by auto
  from subS subNS have subNSS:"sub_closed (\<succeq> \<union> \<succ>)" unfolding sub_closed_def using cs_union_distr[where ?R="\<succeq>" and ?S=\<succ>] by auto
  with nonStrict monoNSS subNSS trs_inclusion have "(rstep R) \<subseteq> (\<succeq> \<union> \<succ>)" by auto
  with `\<forall>i. (t i, t (Suc i)) \<in> rstep R` have "\<forall>i. (t i, t (Suc i)) \<in> (\<succeq> \<union> \<succ>)" by auto
  with compat `SN_elt \<succ> (t 0)` have "\<exists> j. \<forall> i \<ge> j.  (t i, t (Suc i)) \<in> (\<succeq> - \<succ>)" using non_strict_ending by auto
  then obtain j where stepNS:" \<forall> i \<ge> j.  (t i, t (Suc i)) \<in> (\<succeq> - \<succ>)" by auto
  have alli:"\<forall> i \<ge> j.  (t i, t (Suc i)) \<in> (rstep rR)"
  proof
    fix i
    show "i \<ge> j \<longrightarrow> (t i, t (Suc i)) \<in> rstep rR"
    proof
      assume "i \<ge> j"
      from alli have "(t i, t (Suc i)) \<in> rstep R" by auto thm rstep_diff
      with `i \<ge> j` stepNS `(rstep (R - rR)) \<subseteq> \<succ>` have "(t i, t (Suc i)) \<in>  (rstep R - rstep (R - rR))" by auto
      hence "(t i, t (Suc i)) \<in>  rstep (R - (R - rR))" using rstep_diff[where ?R=R and ?R'="R - rR"] by auto
      from `rR \<subseteq> R` have "R - (R - rR) \<subseteq> rR" by auto
      with `(t i, t (Suc i)) \<in>  rstep (R - (R - rR))`  show "(t i, t (Suc i)) \<in> (rstep rR)"  using rstep_mono by auto
    qed
  qed
  from alli have "\<forall> k. (shift t j k, shift t j (Suc k)) \<in> (rstep rR)" by auto
  thus ?thesis  unfolding ideriv_def by auto
qed

end

(* now the implementation starts *)

abbreviation internal_error where "internal_error \<equiv> (
  ''internal error'' +#+ shows_nl +@+
  ''please contact one of the CeTA members'' +#+ shows_nl +@+
  ''http://cl-informatik.uibk.ac.at/software/ceta'' +#+ shows_nl +@+
  ''ceta@informatik.uibk.ac.at'' +#+ shows_nl
)"

fun checkRedPairProc :: "
  (('f,'v)rule \<Rightarrow> shows monad) \<Rightarrow>
  (('f,'v)rule \<Rightarrow> shows monad) \<Rightarrow> 
   shows \<Rightarrow>
   shows \<Rightarrow> 
   ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> shows monad"
where "checkRedPairProc checkSS checkNSS name representation P rP r  = (
  let (Sp::('f,'v)trsL,NSp) = List.partition (\<lambda>foo. isOK(checkSS foo)) P in (
   checkAllM checkNSS r
    \<leftarrow> (\<lambda>s. ''problem when orienting TRS'' +#+ shows_nl +@+ s) 
 \<guillemotright> checkAllM checkNSS NSp
    \<leftarrow> (\<lambda>s. ''problem when orienting DPs'' +#+ shows_nl +@+ s)
 \<guillemotright> checkSubset NSp rP
    \<leftarrow> (\<lambda>s. case checkSS s of Left err \<Rightarrow> err | Right ok \<Rightarrow> internal_error)
) \<leftarrow> (\<lambda>s. (''could not apply the reduction pair processor with the following'' +#+ shows_nl +@+
          representation +@+ shows_nl +@+ s))
)"

fun checkMonoRedPairProc :: "
  shows monad \<Rightarrow>
  (('f,'v)rule \<Rightarrow> shows monad) \<Rightarrow>
  (('f,'v)rule \<Rightarrow> shows monad) \<Rightarrow> 
   shows \<Rightarrow>
   shows \<Rightarrow> 
   ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL  \<Rightarrow>  ('f,'v)trsL  \<Rightarrow> shows monad"
where "checkMonoRedPairProc checkMono checkSS checkNSS  name representation P rP R rR  = ( 
  let (Sp :: ('f,'v) trsL , NSp) = List.partition (\<lambda> foo. isOK(checkSS foo)) P in
  let (Sr :: ('f,'v) trsL , NSr) = List.partition (\<lambda> foo. isOK(checkSS foo)) R in (
   checkMono
    \<leftarrow> (\<lambda>s. ''problem with monotonicity of strict order'' +#+ shows_nl +@+ s)
 \<guillemotright> checkAllM checkNSS NSr
    \<leftarrow> (\<lambda>s. ''problem when orienting TRS'' +#+ shows_nl +@+ s) 
 \<guillemotright> checkAllM checkNSS NSp
    \<leftarrow> (\<lambda>s. ''problem when orienting DPs'' +#+ shows_nl +@+ s)
 \<guillemotright> checkSubset rR R
    \<leftarrow> (\<lambda>s. shows_string ''problem: additional '')
 \<guillemotright> checkSubset NSr rR
    \<leftarrow> (\<lambda>s. case checkSS s of Left err \<Rightarrow> err | Right ok \<Rightarrow> internal_error)
 \<guillemotright> checkSubset NSp rP
    \<leftarrow> (\<lambda> s. case checkSS s of Left err \<Rightarrow> err | Right ok \<Rightarrow> internal_error)
) \<leftarrow> (\<lambda> s. ''could not apply the reduction pair processor with the following'' +#+ shows_nl +@+
          representation +@+ shows_nl +@+ s)
)"

fun checkRuleRemovalProc :: "
  shows monad \<Rightarrow> 
  (('f,'v)rule \<Rightarrow> shows monad) \<Rightarrow>
  (('f,'v)rule \<Rightarrow> shows monad) \<Rightarrow> 
   shows \<Rightarrow>
   shows \<Rightarrow> 
   ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> shows monad"
where "checkRuleRemovalProc checkMono checkSS checkNSS  name representation R rR  = 
( let (Sp :: ('f,'v) trsL , NSp) = List.partition (\<lambda> foo. isOK(checkSS foo)) R in
    (
   checkMono
    \<leftarrow> (\<lambda>s. ''problem with monotonicity of strict order'' +#+ shows_nl +@+ s)
 \<guillemotright> checkAllM checkNSS R
    \<leftarrow> (\<lambda>s. ''problem when orienting rules'' +#+ shows_nl +@+ s) 
 \<guillemotright> checkSubset NSp rR
    \<leftarrow> (\<lambda>s. case checkSS s of Left err \<Rightarrow> err | Right ok \<Rightarrow> internal_error)
 \<guillemotright> checkSubset rR R
    \<leftarrow> (\<lambda>s. case checkSS s of Left err \<Rightarrow> err | Right ok \<Rightarrow> internal_error)
    )
  \<leftarrow> (\<lambda>s. (''could not apply the rule removal processor with the following'' +#+ shows_nl +@+
         representation +@+ shows_nl +@+ s))
 )"

locale redpair_impl = redpair + 
  fixes checkS :: "('f::shows,'v::shows)rule \<Rightarrow> shows monad"
    and checkNS :: "('f,'v)rule \<Rightarrow> shows monad"
  assumes checkSSound: "isOK(checkS rule) \<Longrightarrow> rule \<in> S" 
    and   checkNSSound: "isOK(checkNS rule)  \<Longrightarrow> rule \<in> NS"
begin

lemma check_redpair_proc_sound[simp] :
  shows "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(checkRedPairProc checkS checkNS name representation P rP R))"
unfolding chain_identifying_onlyDP_proc_impl_def
proof ((rule allI)+, rule impI)
  fix P R rP 
  assume check: "isOK (checkRedPairProc checkS checkNS name representation P rP R)"
  have "redPairProc \<succ> \<succeq> (set P) (set R) (set rP) (set R)" 
  proof (cases "List.partition (\<lambda> dp. isOK(checkS dp)) P")
    case (Pair Sp NSp)
    let ?NSp = "set NSp \<inter> set rP"
    from Pair check checkSSound checkNSSound have cond:
      "set P - set NSp \<subseteq> S \<and> 
      ?NSp \<inter> set P \<subseteq> NS \<union> S   \<and> 
      set R \<subseteq> NS     \<and>
      set NSp \<subseteq> set rP"
      by (auto simp: Let_def mem_iff)
    hence id: "set NSp = ?NSp" and stri: "set P - ?NSp \<subseteq> S" by (auto simp: mem_iff)
    with cond show ?thesis by auto
  qed
  with redPairProc_ci show "\<forall> s t \<sigma>. min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i))"
    unfolding chain_identifying_proc_def by auto
qed
end

locale mono_redpair_impl = redpair_impl + 
  fixes checkMono :: "shows monad"
  assumes checkMonoSound : "isOK(checkMono) \<Longrightarrow> one_step_ctxt_closed S"
begin

lemma rule_removal_sound[simp] : 
  shows check: "ci_tt_impl (\<lambda> R rR. isOK(checkRuleRemovalProc checkMono checkS checkNS name representation R rR))"
unfolding ci_tt_impl_def
proof (intro allI, intro impI)
  fix R rR
  assume check: "isOK(checkRuleRemovalProc checkMono checkS checkNS name representation R rR)"
  from check checkMonoSound have ccS: "one_step_ctxt_closed S" by auto
  from check checkSSound checkNSSound have 
    "set R  \<subseteq> NS \<union> S" 
    and "set R - set rR \<subseteq> S" 
    and "set rR \<subseteq> set R" 
   by (auto simp: Let_def mem_iff)
 thus "set rR \<subseteq>ci set R \<and> (\<forall> t. ideriv (set R) t \<longrightarrow> (\<exists> i. ideriv (set rR) (shift t i)))"
   using ruleRemoval[where ?R="set R" and ?rR="set rR"] ci_subsetI[of "set rR" "set R"] ccS by simp
qed

lemma check_mono_redpair_proc_sound[simp] :
  shows "chain_identifying_proc_impl (\<lambda> P R rm rP rR. isOK(checkMonoRedPairProc checkMono checkS checkNS name repr P rP R rR))"
unfolding chain_identifying_proc_impl_def
proof ((rule allI)+, rule impI)
  fix P R rP rR
  assume check: "isOK (checkMonoRedPairProc checkMono checkS checkNS name repr P rP R rR)"
  have mono: "monoRedPairProc \<succ> \<succeq> (set P) (set R) (set rP) (set rR)" 
  proof (cases "List.partition (\<lambda> dp. isOK(checkS dp)) P")
    case (Pair Sp NSp)
    hence splitP:"List.partition (\<lambda> dp. isOK(checkS dp)) P = (Pair Sp NSp)" .
    show ?thesis
    proof (cases "List.partition (\<lambda> r. isOK(checkS r)) R")
      case (Pair Sr NSr)
      hence splitR:"List.partition (\<lambda> r. isOK(checkS r)) R = (Pair Sr NSr)" .
      let ?NSp = "set NSp \<inter> set rP"
      let ?NSr = "set NSr \<inter> set rR"
      from splitP check checkSSound checkNSSound have condP:
      "set P - set NSp \<subseteq> S \<and> 
      ?NSp \<inter> set P \<subseteq> NS \<union> S   \<and> 
      set NSp \<subseteq> set rP"
	by (auto simp: Let_def mem_iff)
      hence idP: "set NSp = ?NSp" and striP: "set P - ?NSp \<subseteq> S" by (auto simp: mem_iff)
      from splitR check checkSSound checkNSSound have condR:
      "set R - set NSr \<subseteq> S \<and> 
      ?NSr \<inter> set R \<subseteq> NS \<union> S   \<and> 
      set NSr \<subseteq> set rR"
	by (auto simp: Let_def mem_iff)
      hence idR: "set NSr = ?NSr" and striR: "set R - ?NSr \<subseteq> S" by (auto simp: mem_iff)
      from check checkMonoSound have mono: "one_step_ctxt_closed S" by (auto simp: Let_def)
      from check condP striP mono idP condR striR idR show ?thesis by auto
    qed
  qed
  show "set rR \<subseteq>ci set R \<and> (\<forall> s t \<sigma>. min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set rR) (shift s i) (shift t i) (shift \<sigma> i)))" (is "?goal (set P) (set R) (set rP) (set rR)")
  proof -
    let ?prop = "\<lambda> P R rP rR. monoRedPairProc S NS P R rP rR \<longrightarrow> ?goal P R rP rR"
    from monoRedPairProc_ci have "\<forall> P R rP rR. ?prop P R rP rR"
      unfolding chain_identifying_proc_def  .
    hence "\<forall> R rP rR. ?prop (set P) R rP rR" ..
    hence "\<forall> rP rR. ?prop (set P) (set R) rP rR" ..
    hence "?prop (set P) (set R) (set rP) (set rR)" using ci_subsetI[of "set rR" "set R"] by force
    with mono show ?thesis by blast
  qed
qed
end

end

