(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory UsableRules imports RedPair Substitution Loop Edg begin


(* perhaps replace poss_list by this function *)
fun possImpl :: "('f,'v)term \<Rightarrow> pos list"
where "possImpl(Var x) = [\<epsilon>]"
    | "possImpl(Fun f ss) = (\<epsilon> # concat (map (\<lambda> (i,ps). map ( op <# i) ps) (zip [0 ..< length ss] (map possImpl ss))))"

lemma possImplSound[simp] : "set (possImpl s) = poss s" 
proof (induct s, simp)
  case (Fun f ss)
  let ?z = "zip [0..<length ss] (map possImpl ss)"
  have len: "length ?z = length ss" by auto
  have "(\<Union>a\<in>set ?z.
         set (split (\<lambda>i. map (op <# i)) a)) = 
       {i <# p |i p. i < length ss \<and> p \<in> poss (ss ! i)}" (is "?l = ?r")
  proof (rule set_ext)
    fix ip
    show "(ip \<in> ?l) = (ip \<in> ?r)"
    proof
      assume "ip \<in> ?l" 
      from this obtain ipI where 
	z: "ipI \<in> set ?z" and 
	ip: "ip \<in> set (split (\<lambda> i. map (op <# i)) ipI)" 
	by auto     
      from z obtain i where i: "i < length ?z" and zi: "?z ! i = ipI" by (force simp: set_conv_nth)      
      with ip Fun show "ip \<in> ?r" by auto
    next
      assume "ip \<in> ?r"
      from this obtain i p where i: "i < length ss" and "p \<in> poss (ss ! i)"
	and ip: "ip = i <# p" by auto
      with Fun have p: "p \<in> set (possImpl (ss ! i))" and iz: "i < length ?z" by auto
      from i have id: "?z ! i = (i, possImpl (ss ! i))" (is "_ = ?ipI") by auto
      from iz have  "?z ! i \<in> set ?z" by (rule nth_mem)
      with id have inZ: "?ipI \<in> set ?z" by auto
      from p have "i <# p \<in> set (split (\<lambda> i. map (op <# i)) ?ipI)" by auto
      with inZ ip show "ip \<in> ?l" by force
    qed
  qed
  with Fun show ?case by simp
qed

declare possImpl.simps[simp del]

fun rewrite :: "('f,'v)trsL \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term list"
where "rewrite R s = concat (map (\<lambda> (l,r) . concat (map (\<lambda> p.
         (case match (s|_p) l of None \<Rightarrow> [] | Some \<sigma> \<Rightarrow> [(ctxt_of_pos_term p s)\<langle>r\<cdot>\<sigma>\<rangle>])
       ) (possImpl s)) ) R)"


lemma rewriteComplete:
  fixes l r::"('f,'v)term"
  assumes "(s,t) \<in> rstep_r_p_s (set R) (l,r) p \<sigma>" and "vars_term l \<supseteq> vars_term r"
  shows "t \<in> set (rewrite R s)" 
proof -
  let ?C = "ctxt_of_pos_term p s"
  from assms have lr: "(l,r) \<in> set R" 
    and p: "p \<in> poss s" 
    and s: "?C\<langle>l \<cdot> \<sigma>\<rangle> = s"
    and t: "?C\<langle>r \<cdot> \<sigma>\<rangle> = t"
    unfolding rstep_r_p_s_def by (auto simp: Let_def)
  from s p have "?C\<langle>l \<cdot> \<sigma>\<rangle> = ?C\<langle>s|_p\<rangle>" by (simp add: ctxt_supt_id)
  hence "l \<cdot> \<sigma> = s|_p" by auto
  from this matchComplete obtain \<delta> where 
    match: "match (s|_p) l = Some \<delta>" and 
           "\<forall> x \<in> vars_term l. getSub \<sigma> x = getSub \<delta> x" by blast
  with assms lr have vars: "\<forall> x \<in> vars_term r. getSub \<sigma> x = getSub \<delta> x" by blast
  have "r \<cdot> \<sigma> = r \<cdot> (\<sigma>|_(vars_term r))" by (rule coincidence_lemma)
  also have "\<dots> = r \<cdot> (\<delta>|_(vars_term r))" 
  proof -
    have "\<sigma>|_(vars_term r) = \<delta>|_(vars_term r)" 
    proof (cases \<sigma>, cases \<delta>)
      fix sigma delta
      assume s: "\<sigma> = Sub sigma" and d: "\<delta> = Sub delta"
      with vars have vars2: "\<forall> x \<in> vars_term r. sigma x = delta x" by auto
      let ?P = "\<lambda> y. (\<lambda> x. if x \<in> vars_term r then sigma x else Var x) y = 
	(\<lambda> x. if x \<in> vars_term r then delta x else Var x) y"
      have "\<And> y. ?P y"	
      proof -
	fix y
	from vars2 show "?P y" by (cases "y \<in> vars_term r", auto)
      qed      
      with s d show ?thesis by auto  
    qed
    thus ?thesis by auto
  qed
  also have "\<dots> = r \<cdot> \<delta>" by (rule coincidence_lemma[symmetric])
  finally  have nearlyDone: "(l,r) \<in> set R \<and> p \<in> poss s \<and> t \<in> set (case match (s|_p) l of None \<Rightarrow> [] | Some \<sigma> \<Rightarrow> [ ?C\<langle>r\<cdot>\<sigma>\<rangle>])" using lr p match t by auto 
  hence "t \<in> set ((\<lambda>(l, r). concat  (map 
     (\<lambda>p. case match (s|_p) l  of None \<Rightarrow> [] | Some \<sigma> \<Rightarrow> [ctxt_of_pos_term p s\<langle>r \<cdot> \<sigma>\<rangle>])
                          (possImpl s)))  (l,r))" (is "t \<in> (?P (l,r))")
    by auto
  with lr have "\<exists> lr \<in> set R. t \<in> (?P lr)" by blast
  thus ?thesis by auto
qed

lemma rewriteSound : assumes "t \<in> set (rewrite R s)"
  shows "(s,t) \<in> rstep (set R)"
proof -
  from assms 
  have "\<exists> l r p.  (l,r) \<in> set R \<and> p \<in> poss s \<and> t \<in> set (case match (s|_p) l of None \<Rightarrow> [] | Some \<sigma> \<Rightarrow> [ (ctxt_of_pos_term p s)\<langle>r\<cdot>\<sigma>\<rangle>])" 
    by (auto simp: mem_iff, blast)
  from this obtain l r p where one:
    "(l,r) \<in> set R \<and> p \<in> poss s \<and> t \<in> set (case match (s|_p) l of None \<Rightarrow> [] | Some \<sigma> \<Rightarrow> [ (ctxt_of_pos_term p s)\<langle>r\<cdot>\<sigma>\<rangle>])"
    by auto
  from this obtain \<sigma> where two: "match (s|_p) l = Some \<sigma> \<and> t \<in> {(ctxt_of_pos_term p s)\<langle>r\<cdot>\<sigma>\<rangle>}"
    by (cases "match (s|_p) l", auto)
  hence match: "l \<cdot> \<sigma> = s|_p" using matchSound by auto 
  with one match one two have "(s,t) \<in> rstep_r_p_s (set R) (l,r) p \<sigma>" unfolding rstep_r_p_s_def by (simp add: Let_def ctxt_supt_id)
  thus "(s,t) \<in> rstep (set R)" using rstep_iff_rstep_r_p_s by blast
qed



fun comb :: "('f \<times> 'f \<times> nat) \<Rightarrow> ('f,'v)term list \<Rightarrow> ('f,'v)term"
where "comb (c,d,_) [] = Fun d []"
   |  "comb (c,d,n) (t # ts) = Fun c (t # comb (c,d,n) ts # replicate n (Fun d []))"

(* here we have one problem, there might be only one variable! *)
(* therefore, we directly create all instances *)
fun ceTrs :: "('f \<times> 'f \<times> nat) \<Rightarrow> ('f,'v)trs"
where "ceTrs (c,d,n) = 
  {(Fun c (t # s # replicate n (Fun d [])), t) | t s. True}
  \<union> {(Fun c (t # s # replicate n (Fun d [])), s) | t s. True}" 


lemma comp_one_step_ctxt_closed : assumes 
  "one_step_ctxt_closed r" and "one_step_ctxt_closed s"
  shows "one_step_ctxt_closed (s O r)"
using assms unfolding one_step_ctxt_closed_def by force

lemma union_one_step_ctxt_closed : assumes 
  "one_step_ctxt_closed r" and "one_step_ctxt_closed s"
  shows "one_step_ctxt_closed (r \<union> s)"
using assms unfolding one_step_ctxt_closed_def by force
  

lemma free_right_rewrite_imp_not_SN:
  fixes l r::"('f,'v)term"
  assumes step: "(t,s) \<in> rstep_r_p_s R (l,r) p \<sigma>" 
  and vars: "\<not> vars_term l \<supseteq> vars_term r" 
  shows "\<not> SN_elt (rstep R) t" 
proof
  let ?C = "ctxt_of_pos_term p t"
  from step have left: "?C\<langle>l \<cdot> \<sigma>\<rangle> = t" (is "?t = t") and right: "?C\<langle>r \<cdot> \<sigma>\<rangle> = s" and pos: "p \<in> poss t"
    and rule: "(l,r) \<in> R"
    unfolding rstep_r_p_s_def by (auto simp: Let_def)
  assume "SN_elt (rstep R) t"
  with left have sn: "SN_elt (rstep R)  ?t" by auto
  have "?t \<unrhd> l \<cdot> \<sigma>" by (rule ctxt_imp_supteqp)
  with sn have "SN_elt (rstep R) (l \<cdot> \<sigma>)" by (rule SN_imp_SN_subt)
  hence snl: "SN_elt (rstep R) l" by (rule SNinstance_imp_SN)
  
  from vars obtain x where xr: "x \<in> vars_term r" and xl: "x \<notin> vars_term l" by auto
  from xr have "r \<unrhd> Var x" (is "_ \<unrhd> ?x") by (rule var_imp_supteqp)
  from this obtain q where q: "q \<in> poss r" and qx: "r|_q = ?x" using supteqp_imp_subt_at by auto
  let ?sig = "subst x l"
  let ?rS = "r \<cdot> \<sigma>"
  from q have "q \<in> poss ?rS" by auto
  from `r \<unrhd> ?x` obtain C where Cr: "C\<langle>?x\<rangle> = r" (is "?r = r") using supteqp_imp_ctxt by auto
  let ?Cs = "C \<cdot>c ?sig"

  let ?sigi = "\<lambda>y. if x = y then l else Var y"
  let ?sl = "(\<lambda>z.(if z \<in> vars_term l then ?sigi z else (Var z)))"
  have sl1: "?sl = getSub \<iota>" proof (rule ext)
    fix z
    show "?sl z = getSub \<iota> z" using xl by (cases "z \<in> vars_term l", auto)
  qed
  have "?sig|_(vars_term l) = Sub ?sl" unfolding subst_def by auto
  with sl1 have "?sig|_(vars_term l) = \<iota>" by auto
  hence lid: "l \<cdot> ?sig = l" using xl coincidence_lemma[where t = l and \<sigma> = ?sig] by auto
  from rule have "(l \<cdot> ?sig, r \<cdot> ?sig) \<in> CS R" by (rule csI)
  with lid Cr[symmetric] have "(l, ?r \<cdot> ?sig) \<in> CS R" by auto
  hence "(\<box>\<langle>l\<rangle>, \<box>\<langle>?r \<cdot> ?sig\<rangle>) \<in> CC (CS R)" by (rule ccI)
  hence "(l, ?r \<cdot> ?sig) \<in> rstep R" by auto
  hence "is_loop ([l], ?Cs, ?sig) R" by (simp add: lid, simp add: subst_def)
  hence "\<not> SN_elt (rstep R) (hd [l])" by (rule is_loop_implies_not_SN_elt)
  with snl show False by auto
qed

lemma ceTrsSound: "t mem ts \<Longrightarrow> (comb cn ts, t) \<in> (rstep (ceTrs cn))^+"
using assms 
proof (induct ts, simp)
  case (Cons s ss)
  show ?case 
  proof (cases cn)
    case (fields c d n)
    from Cons show ?thesis 
    proof (cases "s = t")
      case True 
      with fields have "(comb cn (s # ss),s) \<in> ceTrs cn" by simp
      hence "(comb cn (s # ss),t) \<in> rstep (ceTrs cn)" 
	using rstep_contains_r True by blast
      thus ?thesis using True by force
    next
      case False
      hence ind: "(comb cn ss, t) \<in> (rstep (ceTrs cn))^+" using Cons by auto
      from fields have "(comb cn (s # ss),comb cn ss) \<in> ceTrs cn" by simp
      hence "(comb cn (s # ss),comb cn ss) \<in> rstep (ceTrs cn)" 
	using rstep_contains_r  by blast
      with ind show ?thesis by auto
    qed
  qed
qed

declare ceTrs.simps comb.simps[simp del]


fun iTransCheck :: "('f,'v)trsL \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)term \<Rightarrow> bool"
where "iTransCheck R ur (Fun f ts) = ( \<not> (\<exists> (l,r) \<in> set R - ur. matchCapBelow l R (Fun f ts)))"

declare iTransCheck.simps[simp del]

function iTrans :: "(('f,'v)trsL \<times> ('f,'v)trs \<times> ('f \<times> 'f \<times> nat)) \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term"
where "iTrans _ (Var x) = (Var x)" 
   |  "iTrans (R, ur, cn) (Fun f ts) = (if SN_elt (rstep (set R)) (Fun f ts) then (
         if iTransCheck R ur (Fun f ts) then Fun f (map (iTrans (R, ur, cn)) ts) 
                   else (comb cn) (Fun f (map (iTrans (R, ur, cn)) ts) # map (iTrans (R, ur, cn)) (rewrite R (Fun f ts)))
  )  else Fun f ts)"
by (pat_completeness, auto)


termination iTrans
proof -
  {
    fix R :: "('f,'v)trsL" 
    and ur :: "('f,'v)trs" 
    and cn :: "'f \<times> 'f \<times> nat"
    have "\<forall> t. iTrans_dom ((R,ur,cn),t)"
    proof (clarify)
      fix t
      let ?Ruc = "(R,ur,cn)"
      have "SN (snRelSupt (rstep (set R)))" by (rule SN_snRelRstepSupt)
      hence "wf ((snRelSupt (rstep (set R)))\<inverse>)" by (simp add: SN_iff_wf)
      thus "iTrans_dom (?Ruc, t)"
      proof (induct)
	fix t    
	assume ind: "\<And>s. (s, t) \<in> (snRelSupt (rstep (set R)))\<inverse> \<Longrightarrow> iTrans_dom (?Ruc, s)"
	show "iTrans_dom (?Ruc, t)"
	proof 
	  fix s
	  assume step: "iTrans_rel s (?Ruc, t)"
	  thus "iTrans_dom s"
	  proof
	    fix R' :: "('f,'v)trsL" and ur' :: "('f,'v)trs" and cn' :: "'f \<times> 'f \<times> nat" 
	      and  f :: 'f and ts :: "('f,'v)term list" and t' :: "('f,'v)term"
	    let ?Ruc' = "(R',ur',cn')"
	    assume s: "s = (?Ruc', t')"
	    assume t: "(?Ruc,t) = (?Ruc', Fun f ts)"
	    assume t': "t' \<in> set ts"
	    assume SN: "SN_elt (rstep (set R')) (Fun f ts)"
	    from t' t SN have "t \<rhd> t'" and "SN_elt (rstep (set R)) t" by auto
	    hence "(t',t) \<in> (snRelSupt (rstep (set R)))\<inverse>" unfolding snRelSupt_def supt_def snRel_def by auto
	    with ind s t show "iTrans_dom s" by auto
	  next
	    fix R' :: "('f,'v)trsL" and ur' :: "('f,'v)trs" and cn' :: "'f \<times> 'f \<times> nat" 
	      and  f :: 'f and ts :: "('f,'v)term list" and t' :: "('f,'v)term"
	    let ?Ruc' = "(R',ur',cn')"
	    assume s: "s = (?Ruc', t')"
	    assume t: "(?Ruc,t) = (?Ruc', Fun f ts)"
	    assume t': "t' \<in> set ts"
	    assume SN: "SN_elt (rstep (set R')) (Fun f ts)"
	    from t' t SN have "t \<rhd> t'" and "SN_elt (rstep (set R)) t" by auto
	    hence "(t',t) \<in> (snRelSupt (rstep (set R)))\<inverse>" unfolding snRelSupt_def supt_def snRel_def by auto
	    with ind s t show "iTrans_dom s" by auto
	  next
	    fix R' :: "('f,'v)trsL" and ur' :: "('f,'v)trs" and cn' :: "'f \<times> 'f \<times> nat" 
	      and  f :: 'f and ts :: "('f,'v)term list" and t' :: "('f,'v)term"
	    let ?Ruc' = "(R',ur',cn')"
	    assume s: "s = (?Ruc', t')"
	    assume t: "(?Ruc,t) = (?Ruc', Fun f ts)"
	    assume t': "t' \<in> set (rewrite R' (Fun f ts))"
	    assume SN: "SN_elt (rstep (set R')) (Fun f ts)"
	    from t t' rewriteSound have step: "(t, t') \<in> rstep (set R)" by blast
	    from SN t have "SN_elt (rstep (set R)) t" by auto
	    with step have "(t',t) \<in> (snRelSupt (rstep (set R)))\<inverse>" unfolding snRelSupt_def
	      snRel_def by auto
	    with ind s t show "iTrans_dom s" by auto
	  qed
	qed
      qed
    qed
  }
  thus ?thesis by force
qed


fun iTransSig :: "(('f,'v)trsL \<times> ('f,'v)trs \<times> ('f \<times> 'f \<times> nat)) \<Rightarrow> ('f,'v)sub \<Rightarrow> ('f,'v)sub"
  where "iTransSig params (Sub \<sigma>) = (Sub (\<lambda> x. iTrans params (\<sigma> x)))" 

types 'f af = "'f \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> bool"


fun urClosedT :: "('f,'v)term \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trs \<Rightarrow> bool"
where "urClosedT (Var x) _ _ = True"
   |  "urClosedT (Fun f ts) R ur = ((\<forall> t \<in> set ts. urClosedT t R ur) \<and> 
  (\<forall> (l,r) \<in> set R. matchCapBelow l R (Fun f ts) \<longrightarrow> (l,r) \<in> ur) 
)" 

fun urClosedT_af :: "('f,'v)term \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trs \<Rightarrow> 'f af \<Rightarrow> bool"
where "urClosedT_af (Var x) _ _ _ = True"
   |  "urClosedT_af (Fun f ts) R ur \<pi> = ((\<forall> i < length ts. \<pi> f (length ts) i \<longrightarrow> urClosedT_af (ts ! i) R ur \<pi>) \<and> 
  (\<forall> (l,r) \<in> set R. matchCapBelow l R (Fun f ts) \<longrightarrow> (l,r) \<in> ur) 
)" 

abbreviation urClosed :: "('f,'v)trs \<Rightarrow> ('f,'v)trsL \<Rightarrow> bool"
where "urClosed ur R \<equiv> 
   (\<forall> (l,r) \<in> ur. urClosedT r R ur)"

abbreviation urClosed_af :: "('f,'v)trs \<Rightarrow> ('f,'v)trsL \<Rightarrow> 'f af \<Rightarrow> bool"
where "urClosed_af ur R \<pi> \<equiv> 
   (\<forall> (l,r) \<in> ur. urClosedT_af r R ur \<pi>)"


definition af_compatible :: "'f af \<Rightarrow> ('f,'v)trs \<Rightarrow> bool"
where "af_compatible \<pi> ord \<equiv>
   (\<forall> f bef s t aft. \<pi> f (Suc (length bef + length aft)) (length bef) \<or>
   (Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> ord)"

notation Edg.equiv_class ("[_]")
lemma iTransI : assumes ur: "urClosedT t R ur" and sn: "SN_elt (rstep (set R)) (t \<cdot> \<sigma>)"
  shows "iTrans (R, ur, com) (t \<cdot> \<sigma>) = t \<cdot> (iTransSig (R, ur, com) \<sigma>)"
using assms
proof (induct t, cases \<sigma>, simp)
  case (Fun f ts)
  hence sClosed: "\<And> s. s \<in> set ts \<Longrightarrow> urClosedT s R ur" by auto
  from `SN_elt (rstep (set R)) (Fun f ts \<cdot> \<sigma>)` have sns: "\<And> s. s \<in> set ts \<Longrightarrow> SN_elt (rstep (set R)) (s \<cdot> \<sigma>)" 
    by (simp add: SN_imp_SN_arg)
  from `SN_elt (rstep (set R)) (Fun f ts \<cdot> \<sigma>)` have snt: "SN_elt (rstep (set R)) (Fun f ts)" by (rule SNinstance_imp_SN)
  have "iTrans (R, ur, com) (Fun f ts \<cdot> \<sigma>)
       = iTrans (R, ur, com) (Fun f (map (\<lambda> t. t \<cdot> \<sigma>) ts))" by auto
  also have "\<dots> = Fun f (map (iTrans (R,ur,com)) (map (\<lambda> t. t \<cdot> \<sigma>) ts))"
  proof -
    from `urClosedT (Fun f ts) R ur`
    have inst: "\<forall> (l,r) \<in> set R. matchCapBelow l R (Fun f ts) \<longrightarrow> (l,r) \<in> ur" by auto
    have "iTransCheck R ur (Fun f ts \<cdot> \<sigma>)" (is ?check)
    proof (rule ccontr)
      assume "\<not> ?check"
      from this obtain l r \<delta> where nUsable: "(l,r) \<in> set R - ur" and elem: "l \<cdot> \<delta> \<in> [HFun f (map (\<lambda> t. cap (set R) (t \<cdot> \<sigma>)) ts)]" 
	(is "_ \<in> [?hs]")
	by (auto simp:iTransCheck.simps gcMatch_def)
      have "\<And> t. [cap (set R) (t \<cdot> \<sigma>)] \<subseteq> [cap (set R) t]" by (rule capInstanceSubset)
      hence "[?hs] \<subseteq> [HFun f (map (cap (set R)) ts)]" (is "_ \<subseteq> [?h]") by auto
      with elem have "l \<cdot> \<delta> \<in> [?h]" by auto
      hence "gcMatch ?h l" unfolding gcMatch_def ..
      with inst nUsable show False by auto
    qed
    with Fun show ?thesis by auto
  qed
  also have "\<dots> = Fun f (map (\<lambda> t. iTrans (R,ur,com) (t \<cdot> \<sigma>)) ts)" by simp
  also have "\<dots> = Fun f (map (\<lambda> t. t \<cdot> (iTransSig (R,ur,com) \<sigma>)) ts)" using Fun sns by auto
  also have "\<dots> = Fun f ts \<cdot> iTransSig (R,ur,com) \<sigma>" by auto
  finally show ?case .
qed


lemma af_steps_imp_orient: assumes tran: "trans r" 
  and refl: "refl r"
  and one: "one_step_ctxt_closed r" 
  and len: "length ts = length ss"
  and steps: "\<forall>i<length ts. p i \<longrightarrow> (ts ! i, ss ! i) \<in> r"
  and compat: "\<forall> bef s t aft. ((length ts = Suc (length bef + length aft)) \<longrightarrow> (p (length bef) \<or> (Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> r))"
  shows "(Fun f ts, Fun f ss) \<in> r" 
proof -
  let ?n = "length ts"
  let ?us = "\<lambda> i. take i ss @ (take (?n - i) (List.drop i ts))"
  have "\<And> i. i \<le> ?n \<Longrightarrow> (Fun f ts, Fun f (?us i)) \<in> r"
  proof -
    fix i
    assume "i \<le> ?n"
    thus "(Fun f ts, Fun f (?us i)) \<in> r"
    proof (induct i)
      case 0 thus ?case using refl unfolding refl_on_def by auto
    next
      case (Suc i)
      hence i: "i < length ts" by auto
      from Suc have ind: "(Fun f ts, Fun f (?us i)) \<in> r" by auto
      let ?bef = "take i ss"
      let ?aft = "take (?n - Suc i) (List.drop (Suc i) ts)"
      have cstep: "(Fun f (?bef @ ts ! i # ?aft), Fun f (?bef @ ss ! i # ?aft)) \<in> r" (is "(?l,?r) \<in> r")
      proof (cases "p i")
	case True
	with Suc steps have step: "(ts ! i, ss ! i) \<in> r" by auto
	from one step show ?thesis unfolding one_step_ctxt_closed_def by auto
      next
	case False
	from compat have "length ts = Suc (length ?bef + length ?aft) \<longrightarrow> p (length ?bef) \<or> (?l,?r) \<in> r" by force
	hence compat1: "p (length ?bef) \<or> (?l,?r) \<in> r" using i len by auto
	from i len[symmetric] have "length ?bef = i" using length_take by auto
	with False compat1 show "(?l, ?r) \<in> r" by auto
      qed
      have lId: "Fun f (?us i) = ?l" using len i by (auto simp: drop_Suc_conv_tl)
      have rId: "Fun f (?us (Suc i)) = ?r" using len i by (auto simp: take_Suc_conv_app_nth)
      from cstep have "(Fun f (?us i), Fun f (?us (Suc i))) \<in> r" by (simp only: lId rId)
      with  ind show ?case using tran unfolding trans_def by best
    qed
  qed      
  hence "(Fun f ts, Fun f (?us ?n)) \<in> r" by force
  thus "(Fun f ts, Fun f ss) \<in> r" using len by  auto
qed



locale af_redpair = redpair +
  fixes \<pi> :: "'f af"
  assumes af_compat : "af_compatible \<pi> NS"
begin

lemma iTransI_af : assumes ur: "urClosedT_af t R ur \<pi>" and sn: "SN_elt (rstep (set R)) (t \<cdot> \<sigma>)"
  shows "(t \<cdot> (iTransSig (R, ur, com) \<sigma>), iTrans (R, ur, com) (t \<cdot> \<sigma>)) \<in> NS^*"
using assms
proof (induct t, cases \<sigma>, simp)
  case (Fun f ts)
  hence sClosed: "\<And> i. i < length ts \<and> \<pi> f (length ts) i \<Longrightarrow> urClosedT_af (ts ! i) R ur \<pi>" by auto
  from `SN_elt (rstep (set R)) (Fun f ts \<cdot> \<sigma>)` have sns: "\<And> s. s \<in> set ts \<Longrightarrow> SN_elt (rstep (set R)) (s \<cdot> \<sigma>)" 
    by (simp add: SN_imp_SN_arg)
  from `SN_elt (rstep (set R)) (Fun f ts \<cdot> \<sigma>)` have snt: "SN_elt (rstep (set R)) (Fun f ts)" by (rule SNinstance_imp_SN)
  have "iTrans (R, ur, com) (Fun f ts \<cdot> \<sigma>)
       = iTrans (R, ur, com) (Fun f (map (\<lambda> t. t \<cdot> \<sigma>) ts))" (is "?l = _") by auto
  also have "\<dots> = Fun f (map (iTrans (R,ur,com)) (map (\<lambda> t. t \<cdot> \<sigma>) ts))"
  proof -
    from `urClosedT_af (Fun f ts) R ur \<pi>`
    have inst: "\<forall> (l,r) \<in> set R. matchCapBelow l R (Fun f ts) \<longrightarrow> (l,r) \<in> ur" by auto
    have "iTransCheck R ur (Fun f ts \<cdot> \<sigma>)" (is ?check)
    proof (rule ccontr)
      assume "\<not> ?check"
      from this obtain l r \<delta> where nUsable: "(l,r) \<in> set R - ur" and elem: "l \<cdot> \<delta> \<in> [HFun f (map (\<lambda> t. cap (set R) (t \<cdot> \<sigma>)) ts)]" 
	(is "_ \<in> [?hs]")
	by (auto simp:iTransCheck.simps gcMatch_def)
      have "\<And> t. [cap (set R) (t \<cdot> \<sigma>)] \<subseteq> [cap (set R) t]" by (rule capInstanceSubset)
      hence "[?hs] \<subseteq> [HFun f (map (cap (set R)) ts)]" (is "_ \<subseteq> [?h]") by auto
      with elem have "l \<cdot> \<delta> \<in> [?h]" by auto
      hence "gcMatch ?h l" unfolding gcMatch_def ..
      with inst nUsable show False by auto
    qed
    with Fun show ?thesis by auto
  qed
  finally have eq1: "?l = Fun f (map (\<lambda> t. iTrans (R,ur,com) (t \<cdot> \<sigma>)) ts)" (is "_ = ?r") by simp
  let ?tis = "\<lambda> t. t \<cdot> (iTransSig (R,ur,com) \<sigma>)"
  let ?its = "\<lambda> t. iTrans (R,ur,com) (t \<cdot> \<sigma>)"
  have "(Fun f (map ?tis ts), Fun f (map ?its ts)) \<in> NS^*" 
  proof -
    have ind: "\<forall> i < length ts. \<pi> f (length ts) i \<longrightarrow> (?tis (ts ! i), ?its (ts ! i)) \<in> NS^*"
    proof (rule allI, (rule impI)+)
      fix i
      assume ilen: "i < length ts" and ipi: "\<pi> f (length ts) i"
      with sns have "ts ! i \<in> set ts" and "SN_elt (rstep (set R)) (ts ! i \<cdot> \<sigma>)" by auto
      with Fun sClosed ilen ipi show "(?tis (ts ! i), ?its (ts ! i)) \<in> NS^*" by auto
    qed
    from wmNS have cc: "one_step_ctxt_closed (NS^*)" using rtrancl_one_step_ctxt_closed  by auto
    show ?thesis 
    proof (rule af_steps_imp_orient, rule trans_rtrancl, rule refl_rtrancl, rule cc, simp)
      show "\<forall>i<length (map ?tis ts). \<pi> f (length ts) i \<longrightarrow> (map ?tis ts ! i, map ?its ts ! i) \<in> \<succeq>\<^sup>*" 
	using ind by auto
    next
      show "\<forall>bef s t aft. length (map ?tis ts) = Suc (length bef + length aft) \<longrightarrow> \<pi> f (length ts) (length bef) \<or> (Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> \<succeq>\<^sup>*"
	using af_compat unfolding af_compatible_def by force
    qed
  qed
  also have "Fun f (map ?tis ts) = Fun f ts \<cdot> iTransSig (R,ur,com) \<sigma>" by auto
  finally show ?case using eq1  by auto
qed
end


lemma iTransII : assumes sn: "SN_elt (rstep (set R)) (t \<cdot> \<sigma>)" 
  shows "(iTrans (R, ur, cn) (t \<cdot> \<sigma>), t \<cdot> iTransSig (R, ur, cn) \<sigma>) \<in> (rstep (ceTrs cn))^*" 
  (is "(iTrans ?Ruc _,_) \<in> ?rel")
using sn
proof (induct t)
  case (Var x) thus ?case by (cases \<sigma>, simp)
next
  case (Fun f ts)
  from `SN_elt (rstep (set R)) (Fun f ts \<cdot> \<sigma>)` have sns: "\<And> s. s \<in> set ts \<Longrightarrow> SN_elt (rstep (set R)) (s \<cdot> \<sigma>)"
    by (simp add: SN_imp_SN_arg)
  from `SN_elt (rstep (set R)) (Fun f ts \<cdot> \<sigma>)` have snt: "SN_elt (rstep (set R)) (Fun f ts)" by (rule SNinstance_imp_SN)
  from Fun sns have "\<And> s. s \<in> set ts \<Longrightarrow> (iTrans ?Ruc (s \<cdot> \<sigma>), s \<cdot> iTransSig ?Ruc \<sigma>) \<in> ?rel" by simp
  let ?ts = "map (\<lambda> t. iTrans ?Ruc (t \<cdot> \<sigma>)) ts"
  let ?ss = "map (\<lambda> t. t \<cdot> iTransSig ?Ruc \<sigma>) ts"
  have len: "length ?ts = length ?ss" and other: "\<forall> i. i< length ?ts \<longrightarrow> (?ts ! i, ?ss ! i) \<in> ?rel" using sns Fun  by auto
  from len other have "(Fun f ?ts, Fun f ?ss) \<in> ?rel" using rstep_rtrancl_all_ctxt_closed[where r = "ceTrs cn"] unfolding all_ctxt_closed_def by blast 
  hence nearlyDone: "(Fun f ?ts, Fun f ts \<cdot> iTransSig ?Ruc \<sigma>) \<in> ?rel" (is "(_,?ti) \<in> _") by auto
  show ?case
  proof (cases "iTransCheck R ur (Fun f ts \<cdot> \<sigma>)")
    case True
    with Fun nearlyDone show ?thesis by (simp add: o_def) 
  next
    case False
    with Fun have id: "iTrans ?Ruc (Fun f ts \<cdot> \<sigma>) = 
       (comb cn) (Fun f ?ts # map (iTrans ?Ruc) (rewrite R (Fun f ts \<cdot> \<sigma>)))" (is "?it = (comb cn) ?list") 
      by (simp add: o_def) 
    have "Fun f ?ts mem ?list" by simp
    hence "((comb cn) ?list, Fun f ?ts) \<in> (rstep (ceTrs cn))^+" by (rule ceTrsSound) 
    with id nearlyDone show "(?it, ?ti) \<in> ?rel" by auto
  qed
qed

locale ce_redpair = redpair + 
  fixes n :: nat
    and c :: 'f
    and d :: 'f
  assumes NS_ceCompat: "ceTrs (c,d,n) \<subseteq> \<succeq>"

locale ce_af_redpair = ce_redpair + af_redpair
begin

lemma ceOrient : "(rstep (ceTrs (c,d,n)))^* \<subseteq> NS^*"
  by (rule rtrancl_mono, rule trs_inclusion, rule wmNS, rule subNS, rule NS_ceCompat)

lemma iTransII_af : assumes sn: "SN_elt (rstep (set R)) (t \<cdot> \<sigma>)" 
  shows "(iTrans (R, ur, (c,d,n)) (t \<cdot> \<sigma>), t \<cdot> iTransSig (R, ur, (c,d,n)) \<sigma>) \<in> NS^*"
  using sn iTransII ceOrient by blast
end
  

lemma iTransIII : assumes 
  sn: "SN_elt (rstep (set R)) (Fun f ts)" and
  step: "(Fun f ts,s) \<in> rstep_r_p_s (set R) lr \<epsilon> \<sigma>" and
  ur: "iTransCheck R ur (Fun f ts)" and
  urClosed: "urClosed ur R"
  shows "(iTrans (R,ur,cn) (Fun f ts), iTrans (R,ur,cn) s) \<in> (rstep (ceTrs cn))^* O rstep ur \<and> (Fun f ts,s) \<in> rstep ur"
proof -
  let ?rel = "(rstep (ceTrs cn))^*"
  let ?Ruc = "(R,ur,cn)"
  let ?I = "iTrans ?Ruc"
  let ?Is = "iTransSig ?Ruc \<sigma>"
  from step obtain l r where lr: "lr = (l,r)" and rule: "(l,r) \<in> set R" and left: "l \<cdot> \<sigma> = Fun f ts" and right: "r \<cdot> \<sigma> = s"
    unfolding rstep_r_p_s_def by force
  {
    assume "is_var l \<or> \<not> (vars_term l \<supseteq> vars_term r)"
    hence "\<not> (SN_elt (rstep (set R)) (Fun f ts))" (is ?goal)
    proof
      assume "is_var l" 
      with rule obtain x where "(Var x, r) \<in> set R" by (cases l, auto)
      thus ?thesis by (rule left_var_imp_not_SN) 
    next
      from step lr have step2: "(Fun f ts,s) \<in> rstep_r_p_s (set R) (l,r) \<epsilon> \<sigma>" by simp
      assume "\<not> vars_term r \<subseteq> vars_term l"
      with step2 show ?thesis by (rule free_right_rewrite_imp_not_SN)
    qed
    with sn have False by auto
  }
  hence "\<not> (is_var l)" and vars: "vars_term l \<supseteq> vars_term r" by auto
  from this obtain g ss where l: "l = Fun g ss" by (cases l, auto)
  from ur have nmatch: "\<forall> (l,r) \<in> set R - ur. \<not> matchCapBelow l R (Fun f ts)" by (simp only: iTransCheck.simps, blast)
  from l left have fg: "f = g" and tsss: "ts = map (\<lambda> s. s \<cdot> \<sigma>) ss" (is "_ = ?ts") and nvar: "\<not> (is_var l)" by auto
  have "\<forall> i. ss ! i \<cdot> \<sigma> \<in> [cap (set R) (ss ! i \<cdot> \<sigma>)]"
  proof 
    fix i
    show "ss ! i \<cdot> \<sigma> \<in> [cap (set R) (ss ! i \<cdot> \<sigma>)]"
      by (rule capSound[where s = "ss ! i \<cdot> \<sigma>" and t = "ss ! i \<cdot> \<sigma>" and \<sigma> = \<iota>], auto) 
  qed
  hence "matchCapBelow (Fun f ss) R (Fun f (map (\<lambda> s. s \<cdot> \<sigma>) ss))"
    by (simp add: gcMatch_def, blast)
  with l fg tsss rule nmatch have urule: "(l,r) \<in> ur" by blast

  with left[symmetric] right[symmetric] csI[where \<sigma> = \<sigma>] ccI[where C = Hole] 
  have urStep: "(Fun f ts, s) \<in> rstep ur" by force

  from urule urClosed have fr: "urClosedT r R ur" by auto
  from left sn have "SN_elt(rstep (set R)) (l \<cdot> \<sigma>)" by auto
  with iTransII left have "(?I (l \<cdot> \<sigma>), l \<cdot> ?Is) \<in> ?rel" by blast
  with left have one: "(?I (Fun f ts), l \<cdot> ?Is) \<in> ?rel" by auto
  from step lr have "(Fun f ts, s) \<in> rstep (set R)" 
    using rstep_iff_rstep_r_p_s by blast
  with right sn step_preserves_SN[where R = "rstep(set R)"] have sns: "SN_elt (rstep (set R)) (r \<cdot> \<sigma>)" by auto
  with fr  have "?I (r \<cdot> \<sigma>) = r \<cdot> ?Is" by (rule iTransI)
  with right have idRight: "r \<cdot> ?Is = ?I s" by auto
  from urule have "(l \<cdot> ?Is, r \<cdot> ?Is) \<in> rstep_r_p_s ur (l,r) \<epsilon> ?Is"
    unfolding rstep_r_p_s_def by auto
  hence "(l \<cdot> ?Is, r \<cdot> ?Is) \<in> rstep ur" using rstep_iff_rstep_r_p_s by blast
  with idRight have two: "(l \<cdot> ?Is, ?I s) \<in> rstep ur" by auto
  from one two urStep  show ?thesis by auto
qed

context ce_af_redpair
begin
lemma iTransIII_af : assumes 
  sn: "SN_elt (rstep (set R)) (Fun f ts)" and
  step: "(Fun f ts,s) \<in> rstep_r_p_s (set R) lr \<epsilon> \<sigma>" and
  ur: "iTransCheck R ur (Fun f ts)" and
  urClosed: "urClosed_af ur R \<pi>" and
  orient: "ur \<subseteq> NS"
  shows "(iTrans (R,ur,(c,d,n)) (Fun f ts), iTrans (R,ur,(c,d,n)) s) \<in> NS^*"
proof -
  let ?Ruc = "(R,ur,(c,d,n))"
  let ?I = "iTrans ?Ruc"
  let ?Is = "iTransSig ?Ruc \<sigma>"
  from step obtain l r where lr: "lr = (l,r)" and rule: "(l,r) \<in> set R" and left: "l \<cdot> \<sigma> = Fun f ts" and right: "r \<cdot> \<sigma> = s"
    unfolding rstep_r_p_s_def by force
  {
    assume "is_var l \<or> \<not> (vars_term l \<supseteq> vars_term r)"
    hence "\<not> (SN_elt (rstep (set R)) (Fun f ts))" (is ?goal)
    proof
      assume "is_var l" 
      with rule obtain x where "(Var x, r) \<in> set R" by (cases l, auto)
      thus ?thesis by (rule left_var_imp_not_SN) 
    next
      from step lr have step2: "(Fun f ts,s) \<in> rstep_r_p_s (set R) (l,r) \<epsilon> \<sigma>" by simp
      assume "\<not> vars_term r \<subseteq> vars_term l"
      with step2 show ?thesis by (rule free_right_rewrite_imp_not_SN)
    qed
    with sn have False by auto
  }
  hence "\<not> (is_var l)" and vars: "vars_term l \<supseteq> vars_term r" by auto
  from this obtain g ss where l: "l = Fun g ss" by (cases l, auto)
  from ur have nmatch: "\<forall> (l,r) \<in> set R - ur. \<not> matchCapBelow l R (Fun f ts)" by (simp only: iTransCheck.simps, blast)
  from l left have fg: "f = g" and tsss: "ts = map (\<lambda> s. s \<cdot> \<sigma>) ss" (is "_ = ?ts") and nvar: "\<not> (is_var l)" by auto
  have "\<forall> i. ss ! i \<cdot> \<sigma> \<in> [cap (set R) (ss ! i \<cdot> \<sigma>)]"
  proof 
    fix i
    show "ss ! i \<cdot> \<sigma> \<in> [cap (set R) (ss ! i \<cdot> \<sigma>)]"
      by (rule capSound[where s = "ss ! i \<cdot> \<sigma>" and t = "ss ! i \<cdot> \<sigma>" and \<sigma> = \<iota>], auto) 
  qed
  hence "matchCapBelow (Fun f ss) R (Fun f (map (\<lambda> s. s \<cdot> \<sigma>) ss))"
    by (simp add: gcMatch_def, blast)
  with l fg tsss rule nmatch have urule: "(l,r) \<in> ur" by blast

  with left[symmetric] right[symmetric] csI[where \<sigma> = \<sigma>] ccI[where C = Hole] 
  have urStep: "(Fun f ts, s) \<in> rstep ur" by force
  have orientUr: "rstep ur \<subseteq> NS" by (rule trs_inclusion, rule wmNS, rule subNS, rule orient)
  with urStep have nsStep: "(Fun f ts, s) \<in> NS" by auto

  from urule urClosed have fr: "urClosedT_af r R ur \<pi>" by auto
  from left sn have "SN_elt (rstep (set R)) (l \<cdot> \<sigma>)" by auto
  with iTransII_af left have "(?I (l \<cdot> \<sigma>), l \<cdot> ?Is) \<in> NS^*" by blast
  with left have one: "(?I (Fun f ts), l \<cdot> ?Is) \<in> NS^*" by auto
  from step lr have "(Fun f ts, s) \<in> rstep (set R)" 
    using rstep_iff_rstep_r_p_s by blast
  with right sn step_preserves_SN[where R = "rstep(set R)"] have sns: "SN_elt (rstep (set R)) (r \<cdot> \<sigma>)" by auto
  with fr have "(r \<cdot> ?Is, ?I (r \<cdot> \<sigma>)) \<in> NS^*" by (rule iTransI_af)
  with right have idRight: "(r \<cdot> ?Is,?I s) \<in> NS^*" by auto
  from urule have "(l \<cdot> ?Is, r \<cdot> ?Is) \<in> rstep_r_p_s ur (l,r) \<epsilon> ?Is"
    unfolding rstep_r_p_s_def by auto
  hence "(l \<cdot> ?Is, r \<cdot> ?Is) \<in> rstep ur" using rstep_iff_rstep_r_p_s by blast
  with idRight orientUr have two: "(l \<cdot> ?Is, ?I s) \<in> NS^*" by auto
  from one two  show ?thesis by auto
qed
end


lemma iTransIV : assumes 
  sn: "SN_elt (rstep (set R)) (Fun f ts)" and
  step: "(Fun f ts,s) \<in> rstep (set R)" and
  ur: "\<not> iTransCheck R ur (Fun f ts)" 
  shows "(iTrans (R,ur,cn) (Fun f ts), iTrans (R,ur,cn) s) \<in> (rstep (ceTrs cn))^+" (is "_ \<in> ?rel")
proof -
  let ?I = "iTrans (R,ur,cn)"
  from ur sn have id: "?I (Fun f ts) = (comb cn) (Fun f (map ?I ts) # map ?I (rewrite R (Fun f ts)))" (is "_ = (comb cn) ?list") 
    by auto
  from step rstep_iff_rstep_r_p_s 
  obtain l r p \<sigma> where step2: "(Fun f ts,s) \<in> rstep_r_p_s (set R) (l,r) p \<sigma>" by blast
  with free_right_rewrite_imp_not_SN sn have "vars_term l \<supseteq> vars_term r" by blast
  with step2 have "s \<in> set (rewrite R (Fun f ts))" by (rule rewriteComplete)
  hence "(?I s) mem ?list" by (auto simp: mem_iff)
  hence "(comb cn ?list, ?I s) \<in> ?rel" by (rule ceTrsSound)
  with id show ?thesis by simp
qed

context ce_af_redpair
begin

lemma iTransIV_af : assumes 
  sn: "SN_elt (rstep (set R)) (Fun f ts)" and
  step: "(Fun f ts,s) \<in> rstep (set R)" and
  ur: "\<not> iTransCheck R ur (Fun f ts)" 
  shows "(iTrans (R,ur,(c,d,n)) (Fun f ts), iTrans (R,ur,(c,d,n)) s) \<in> NS^*" (is "?ii \<in> NS^*")
proof -
  from sn step ur have "?ii \<in> (rstep (ceTrs (c,d,n)))^+" by (rule iTransIV)
  with ceOrient show ?thesis by auto
qed
end

  
lemma iTransV : assumes 
  sn: "SN_elt (rstep (set R)) t" and
  step: "(t,s) \<in> rstep (set R)" and
  urClosed: "urClosed ur R"
  shows "(iTrans (R,ur,cn) t, iTrans (R,ur,cn) s) \<in> (rstep (ceTrs cn))^* O rstep ur \<and> (t,s) \<in> rstep ur \<or>
         (iTrans (R,ur,cn) t, iTrans (R,ur,cn) s) \<in> (rstep (ceTrs cn))^+"
    (is "_ \<in> ?R \<and> _ \<in> ?Inter \<or> _ \<in> ?Ce")
proof -
  have cc1: "one_step_ctxt_closed ((rstep (ceTrs cn))^*)"
    using one_step_ctxt_closed_rstep[where r = "ceTrs cn"]
    and rtrancl_one_step_ctxt_closed by auto 
  from one_step_ctxt_closed_rstep[where r = ur] this
  have ccR: "one_step_ctxt_closed ?R" by (rule comp_one_step_ctxt_closed)
  from one_step_ctxt_closed_rstep[where r = "ceTrs cn"] cc1
  have ccC: "one_step_ctxt_closed ?Ce" 
    by (simp only: trancl_to_rtrancl, rule comp_one_step_ctxt_closed)
  let ?I = "iTrans (R,ur,cn)"
  let ?goal = "\<lambda> t s. (?I t, ?I s) \<in> ?R \<and> (t,s) \<in> rstep ur \<or>
         (?I t, ?I s) \<in> ?Ce"
  from step sn show "?goal t s"
  proof induct 
    fix t s C
    assume "(t,s) \<in> CS (set R)" and "SN_elt (rstep (set R)) C\<langle>t\<rangle>"
    thus "?goal C\<langle>t\<rangle> C\<langle>s\<rangle>"
    proof (induct C)
      case Hole
      thus ?case
      proof (induct)	
	fix l r \<sigma>
	assume lr: "(l,r) \<in> set R" and sn: "SN_elt (rstep (set R)) \<box>\<langle>l \<cdot> \<sigma>\<rangle>" (is "SN_elt _ ?ls")
	from lr have stepp: "(?ls, \<box>\<langle>r \<cdot> \<sigma>\<rangle>) \<in> rstep_r_p_s (set R) (l,r) \<epsilon> \<sigma>" (is "(_,?rs) \<in> _")
	  unfolding rstep_r_p_s_def by auto
	hence step: "(?ls, ?rs) \<in> rstep (set R)" using rstep_iff_rstep_r_p_s by blast
	with sn have "\<not> (is_var ?ls)" using var_rewrite_imp_not_SN by blast
	from this obtain f ls where ls: "?ls = Fun f ls" by (cases ?ls, auto)
	with sn step stepp have snls: "SN_elt (rstep (set R)) (Fun f ls)" and 
	  stepls: "(Fun f ls, ?rs) \<in> rstep (set R)"  and
	  steppls: "(Fun f ls, ?rs) \<in> rstep_r_p_s (set R) (l,r) \<epsilon> \<sigma>" by auto 
	show "?goal ?ls ?rs"
	proof (cases "iTransCheck R ur (Fun f ls)")
	  case True
	  from snls steppls this urClosed
	  have "(?I (Fun f ls), ?I ?rs) \<in> (rstep (ceTrs cn))^* O rstep ur \<and> (Fun f ls, ?rs) \<in> rstep ur" by (rule iTransIII)
	  with ls show ?thesis by auto
	next
	  case False
	  from snls stepls this 
	  have "(?I (Fun f ls), ?I ?rs) \<in> (rstep (ceTrs cn))^+" by (rule iTransIV)
	  with ls show ?thesis by (auto simp: trancl_to_rtrancl)
	qed
      qed
    next
      case (More f bef D aft)
      hence snbta: "SN_elt (rstep (set R)) (Fun f (bef @ D\<langle>t\<rangle> # aft))" (is "SN_elt _ (Fun f ?bta)") by auto
      let ?Dt = "D\<langle>t\<rangle>"
      let ?Ds = "D\<langle>s\<rangle>"
      let ?bsa = "bef @ ?Ds # aft"
      let ?C = "More f bef D aft"
      from More ccI[where C = ?C] have "(?C\<langle>t\<rangle>, ?C\<langle>s\<rangle>) \<in> rstep (set R)" by auto
      hence step: "(Fun f ?bta, Fun f ?bsa) \<in> rstep (set R)" by auto
      show ?case 
      proof (cases "iTransCheck R ur (Fun f ?bta)")
	case False
	from snbta step False 
	have "(?I (Fun f ?bta), ?I (Fun f ?bsa)) \<in> (rstep (ceTrs cn))^+" by (rule iTransIV)
	thus ?thesis by (auto simp: trancl_to_rtrancl)
      next
	case True
	have "?Dt \<in> set ?bta" by auto 
	with snbta have "SN_elt (rstep (set R)) ?Dt" by (rule SN_imp_SN_arg)
	with More have "?goal ?Dt ?Ds" by auto
	hence ind: "(Fun f (map ?I bef @ ?I ?Dt # map ?I aft), 
	        Fun f (map ?I bef @ ?I ?Ds # map ?I aft)) 
	  \<in> ?R \<and> (Fun f ?bta, Fun f ?bsa) \<in> rstep ur \<or> 
	  (Fun f (map ?I bef @ ?I ?Dt # map ?I aft), 
	        Fun f (map ?I bef @ ?I ?Ds # map ?I aft)) \<in> ?Ce"
	  (is "(?Il, ?Ir) \<in> _ \<and> _ \<or> _") 
	proof
	  assume "(?I ?Dt, ?I ?Ds) \<in> ?R \<and> (?Dt, ?Ds) \<in> rstep ur"
	  with ccR one_step_ctxt_closed_rstep[where r = ur] show ?thesis unfolding one_step_ctxt_closed_def by auto
	next
	  assume "(?I ?Dt, ?I ?Ds) \<in> ?Ce"
	  with ccC show ?thesis unfolding one_step_ctxt_closed_def by auto
	qed

	from True snbta have left: "?I (Fun f ?bta) = ?Il" by auto

	from step snbta have snbsa: "SN_elt (rstep (set R)) (Fun f ?bsa)" by (rule step_preserves_SN)
	from More ccI[where C = D] have "(?Dt,?Ds) \<in> rstep (set R)" by auto
	hence subset: "[cap (set R) ?Ds] \<subseteq> [cap (set R) ?Dt]" by (rule capRewrite)		
	have "iTransCheck R ur (Fun f ?bsa)"
	proof (simp add: iTransCheck.simps, rule ballI) 
	  fix lr
	  assume lr: "lr \<in> set R - ur"
	  show "\<not> (\<lambda>(l, r). gcMatch (HFun f (map (cap (set R)) bef @ cap (set R) ?Ds # map (cap (set R)) aft)) l) lr"
	  proof (cases lr)
	    case (Pair l r)
	    from lr Pair have "(l,r) \<in> set R" by auto
	    from Pair lr have nMatch: 
	      "\<not> gcMatch (HFun f (map (cap (set R)) bef @ cap (set R) ?Dt # map (cap (set R)) aft)) l" (is "\<not> gcMatch ?cbta l") 
	      using True by (auto simp: iTransCheck.simps)
	    {
	      assume "gcMatch (HFun f (map (cap (set R)) bef @ cap (set R) ?Ds # map (cap (set R)) aft)) l" 
	      from this subset have "gcMatch ?cbta l" by (rule gcMatchBelow)	    
	      with nMatch have False ..
	    }
	    with lr Pair show ?thesis by auto 
	  qed
	qed
	with snbsa  have "?I (Fun f ?bsa) = ?Ir" by auto
	
	with left ind show ?thesis by (simp)
      qed
    qed
  qed
qed

context ce_af_redpair 
begin
  
lemma iTransV_af : assumes 
  sn: "SN_elt (rstep (set R)) t" and
  step: "(t,s) \<in> rstep (set R)" and
  urClosed: "urClosed_af ur R \<pi>" and
  orient : "ur \<subseteq> NS"
  shows "(iTrans (R,ur,(c,d,n)) t, iTrans (R,ur,(c,d,n)) s) \<in> NS^*"
proof -
  have cc: "one_step_ctxt_closed (NS^*)" 
    using wmNS and rtrancl_one_step_ctxt_closed by auto 
  let ?I = "iTrans (R,ur,(c,d,n))"
  let ?goal = "\<lambda> t s. (?I t, ?I s) \<in> NS^*"
  from step sn show "?goal t s"
  proof induct 
    fix t s C
    assume "(t,s) \<in> CS (set R)" and "SN_elt (rstep (set R)) C\<langle>t\<rangle>"
    thus "?goal C\<langle>t\<rangle> C\<langle>s\<rangle>"
    proof (induct C)
      case Hole
      thus ?case
      proof (induct)	
	fix l r \<sigma>
	assume lr: "(l,r) \<in> set R" and sn: "SN_elt (rstep (set R)) \<box>\<langle>l \<cdot> \<sigma>\<rangle>" (is "SN_elt _ ?ls")
	from lr have stepp: "(?ls, \<box>\<langle>r \<cdot> \<sigma>\<rangle>) \<in> rstep_r_p_s (set R) (l,r) \<epsilon> \<sigma>" (is "(_,?rs) \<in> _")
	  unfolding rstep_r_p_s_def by auto
	hence step: "(?ls, ?rs) \<in> rstep (set R)" using rstep_iff_rstep_r_p_s by blast
	with sn have "\<not> (is_var ?ls)" using var_rewrite_imp_not_SN by blast
	from this obtain f ls where ls: "?ls = Fun f ls" by (cases ?ls, auto)
	with sn step stepp have snls: "SN_elt (rstep (set R)) (Fun f ls)" and 
	  stepls: "(Fun f ls, ?rs) \<in> rstep (set R)"  and
	  steppls: "(Fun f ls, ?rs) \<in> rstep_r_p_s (set R) (l,r) \<epsilon> \<sigma>" by auto 
	show "?goal ?ls ?rs"
	proof (cases "iTransCheck R ur (Fun f ls)")
	  case True
	  from snls steppls this urClosed orient
	  have "(?I (Fun f ls), ?I ?rs) \<in> NS^*" by (rule iTransIII_af)
	  with ls show ?thesis by auto
	next
	  case False
	  from snls stepls this 
	  have "(?I (Fun f ls), ?I ?rs) \<in> NS^*" by (rule iTransIV_af)
	  with ls show ?thesis by auto
	qed
      qed
    next
      case (More f bef D aft)
      hence snbta: "SN_elt (rstep (set R)) (Fun f (bef @ D\<langle>t\<rangle> # aft))" (is "SN_elt _ (Fun f ?bta)") by auto
      let ?Dt = "D\<langle>t\<rangle>"
      let ?Ds = "D\<langle>s\<rangle>"
      let ?bsa = "bef @ ?Ds # aft"
      let ?C = "More f bef D aft"
      from More ccI[where C = ?C] have "(?C\<langle>t\<rangle>, ?C\<langle>s\<rangle>) \<in> rstep (set R)" by auto
      hence step: "(Fun f ?bta, Fun f ?bsa) \<in> rstep (set R)" by auto
      show ?case 
      proof (cases "iTransCheck R ur (Fun f ?bta)")
	case False
	from snbta step False 
	have "(?I (Fun f ?bta), ?I (Fun f ?bsa)) \<in> NS^*" by (rule iTransIV_af)
	thus ?thesis by simp
      next
	case True
	have "?Dt \<in> set ?bta" by auto 
	with snbta have "SN_elt (rstep (set R)) ?Dt" by (rule SN_imp_SN_arg)
	with More have "?goal ?Dt ?Ds" by auto
	hence ind: "(Fun f (map ?I bef @ ?I ?Dt # map ?I aft), 
	        Fun f (map ?I bef @ ?I ?Ds # map ?I aft)) 
	  \<in> NS^*"
	  (is "(?Il, ?Ir) \<in> _ ") using cc unfolding one_step_ctxt_closed_def by auto 

	from True snbta have left: "?I (Fun f ?bta) = ?Il" by auto

	from step snbta have snbsa: "SN_elt (rstep (set R)) (Fun f ?bsa)" by (rule step_preserves_SN)
	from More ccI[where C = D] have "(?Dt,?Ds) \<in> rstep (set R)" by auto
	hence subset: "[cap (set R) ?Ds] \<subseteq> [cap (set R) ?Dt]" by (rule capRewrite)		
	have "iTransCheck R ur (Fun f ?bsa)"
	proof (simp add: iTransCheck.simps, rule ballI) 
	  fix lr
	  assume lr: "lr \<in> set R - ur"
	  show "\<not> (\<lambda>(l, r). gcMatch (HFun f (map (cap (set R)) bef @ cap (set R) ?Ds # map (cap (set R)) aft)) l) lr"
	  proof (cases lr)
	    case (Pair l r)
	    from lr Pair have "(l,r) \<in> set R" by auto
	    from Pair lr have nMatch: 
	      "\<not> gcMatch (HFun f (map (cap (set R)) bef @ cap (set R) ?Dt # map (cap (set R)) aft)) l" (is "\<not> gcMatch ?cbta l") 
	      using True by (auto simp: iTransCheck.simps)
	    {
	      assume "gcMatch (HFun f (map (cap (set R)) bef @ cap (set R) ?Ds # map (cap (set R)) aft)) l" 
	      from this subset have "gcMatch ?cbta l" by (rule gcMatchBelow)	    
	      with nMatch have False ..
	    }
	    with lr Pair show ?thesis by auto 
	  qed
	qed
	with snbsa  have "?I (Fun f ?bsa) = ?Ir" by auto
	
	with left ind show ?thesis by (simp)
      qed
    qed
  qed
qed
end	
no_notation Edg.equiv_class ("[_]")	

lemma iTransVI : assumes 
  sn: "SN_elt (rstep (set R)) t" and
  step: "(t,s) \<in> (rstep (set R))^*" and
  urClosed: "urClosed ur R"
  shows "(iTrans (R,ur,cn) t, iTrans (R,ur,cn) s) \<in> (rstep (ur \<union> ceTrs cn))^* \<and>
  ((iTrans (R,ur,cn) t, iTrans (R,ur,cn) s) \<in> 
  (rstep (ur \<union> ceTrs cn))^* O
rstep (ceTrs cn) O 
(rstep (ur \<union> ceTrs cn))^*
\<or> (t,s) \<in> (rstep ur)^*)" (is "?goal t s")
using step 
proof (induct rule: rtrancl_induct, simp)
  fix s u
  let ?urc = "rstep (ur \<union> ceTrs cn)"
  let ?ur = "rstep ur"
  let ?c = "rstep (ceTrs cn)"
  let ?R = "?urc^* O ?c O ?urc^*"
  let ?I = "iTrans (R,ur,cn)"
  assume steps: "(t, s) \<in> (rstep (set R))^*" and 
    isteps: "?goal t s" and
    step: "(s,u) \<in> rstep (set R)"
  from steps sn have "SN_elt (rstep (set R)) s" 
    by (rule rtrancl_step_preserves_SN[where R = "rstep(set R)"])
  from this step urClosed have isteps2: "(?I s, ?I u) \<in> ?c^* O ?ur \<and> (s,u) \<in> ?ur \<or> (?I s, ?I u) \<in> ?c^+"
    by (rule iTransV)
  have "?c \<subseteq> ?urc" using rstep_mono by blast
    hence "?c^* \<subseteq> ?urc^*" by (rule rtrancl_mono)
  have "?ur \<subseteq> ?urc" using rstep_mono by blast
  with `?c \<subseteq> ?urc` have "(?c^*) O ?ur \<subseteq> (?urc^*) O ?urc" using rtrancl_mono[where r = ?c and s = ?urc] by blast
  also have "\<dots> \<subseteq> ?urc^*" by auto
  finally have one: "?c^* O ?ur \<subseteq> ?urc^*" . 
  from `?c \<subseteq> ?urc` have "?c^+ \<subseteq> ?urc^+" using trancl_mono by auto
  also have "\<dots> \<subseteq> ?urc^*" by auto
  finally have two: "?c^+ \<subseteq> ?urc^*" by auto
  with one have subset: "?c^* O ?ur \<union> ?c^+ \<subseteq> ?urc^*" by auto
  
  with isteps2 isteps have "(?I s, ?I u) \<in> ?urc^*" and "(?I t, ?I s) \<in> ?urc^*" by auto
  hence oneGoal: "(?I t, ?I u) \<in> ?urc^*" by auto
  thus "?goal t u" 
  proof (rule conjI) 
    from isteps2
    show "(?I t, ?I u) \<in> ?R \<or> (t,u) \<in> ?ur^*"
    proof
      assume "(?I s, ?I u) \<in> ?c^+"
      with `?c^* \<subseteq> ?urc^*` have "(?I s, ?I u) \<in> ?urc^* O ?c" by (auto simp: trancl_to_rtrancl)
      from this obtain z where "(?I s, z) \<in> ?urc^*" and "(z, ?I u) \<in> ?c" by blast
      have "(?I t, ?I u) \<in> ?R"
      proof (intro rel_compI)
	      show "(?I t, z) \<in> ?urc^*"
        using `(?I t, ?I s) \<in> ?urc^*` `(?I s, z) \<in> ?urc^*` by auto
      next
    	  show "(z, ?I u) \<in> ?c" by fact
    	next
	      show "(?I u, ?I u) \<in> ?urc^*" by auto
  	  qed thus ?thesis ..
    next
      assume "(?I s, ?I u) \<in> ?c^* O ?ur \<and> (s,u) \<in> ?ur"
      hence su: "(s,u) \<in> ?ur" ..
      from isteps have "(?I t, ?I s) \<in> ?R \<or> (t,s) \<in> ?ur^*" ..
      thus ?thesis
      proof
      	assume "(t,s) \<in> ?ur^*"
  	    with su have "(t,u) \<in> ?ur^*" by auto
  	    thus ?thesis .. 
      next
  	    assume "(?I t, ?I s) \<in> ?R"
  	    with `(?I s, ?I u) \<in> ?urc^*` have "(?I t, ?I u) \<in> ?R" by force
    	  thus ?thesis ..
      qed
    qed
  qed
qed 

context ce_af_redpair 
begin

lemma iTransVI_af : assumes 
  sn: "SN_elt (rstep (set R)) t" and
  step: "(t,s) \<in> (rstep (set R))^*" and
  urClosed: "urClosed_af ur R \<pi>" and
  orient: "ur \<subseteq> NS"
  shows "(iTrans (R,ur,(c,d,n)) t, iTrans (R,ur,(c,d,n)) s) \<in> NS^*" (is "?goal t s")
using step 
proof (induct rule: rtrancl_induct, simp)
  fix s u
  let ?I = "iTrans (R,ur,(c,d,n))"
  assume steps: "(t, s) \<in> (rstep (set R))^*" and 
    isteps: "?goal t s" and
    step: "(s,u) \<in> rstep (set R)"
  from steps sn have "SN_elt (rstep (set R)) s" 
    by (rule rtrancl_step_preserves_SN[where R = "rstep(set R)"])
  from this step urClosed orient have isteps2: "(?I s, ?I u) \<in> NS^*" 
    by (rule iTransV_af)
  
  with isteps show "?goal t u" by auto
qed
end

declare iTrans.simps ceTrs.simps[simp del]    

abbreviation urPClosed :: "('f,'v)trsL \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> bool" 
  where "urPClosed R ur P \<equiv> \<forall> st \<in> P. urClosedT (snd st) R ur"

abbreviation urPClosed_af :: "('f,'v)trsL \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> 'f af \<Rightarrow> bool" 
  where "urPClosed_af R ur P \<pi> \<equiv> \<forall> st \<in> P. urClosedT_af (snd st) R ur \<pi>"

context ce_redpair
begin
lemma iTransSound : assumes
      chain: "min_ichain (P, set R) s t \<sigma>"
  and urClosed: "urClosed ur R"
  and urPClosed: "urPClosed R ur P"
  shows  "ichain (P, ur \<union> ceTrs cn) s t (\<lambda> i. iTransSig (R,ur,cn) (\<sigma> i)) \<and> 
        (\<forall> i. (t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (rstep ur)^* \<or> 
         (t i \<cdot> iTransSig (R,ur,cn) (\<sigma> i),
          s (Suc i) \<cdot> iTransSig (R,ur,cn) (\<sigma> (Suc i))) \<in> 
          (rstep (ur \<union> ceTrs cn))^* O (rstep (ceTrs cn)) O (rstep (ur \<union> ceTrs cn))^*)"
proof -
  let ?ss = "\<lambda> i. s i \<cdot> \<sigma> i"
  let ?ts = "\<lambda> i. t i \<cdot> \<sigma> i"
  let ?I = "iTrans (R, ur, cn)"
  let ?Is = "\<lambda> i. (iTransSig (R,ur,cn) (\<sigma> i))"
  let ?sIsig = "\<lambda> i. (s i) \<cdot> ?Is i"
  let ?tIsig = "\<lambda> i. (t i) \<cdot> ?Is i"
  let ?ce = "rstep (ceTrs cn)"
  let ?urc = "rstep (ur \<union> ceTrs cn)"
  let ?ur = "rstep ur"
  let ?R = "?urc^* O ?ce O ?urc^*"
  from chain have allP: "\<forall> i. (s i, t i) \<in> P"
    and steps: "\<forall> i. (?ts i, ?ss (Suc i)) \<in> (rstep (set R))^*"
    and snt: "\<forall> i. SN_elt (rstep (set R)) (?ts i)" by (auto simp: minimal_cond_def)
  have main: "\<forall> i. (?tIsig i, ?sIsig (Suc i)) \<in> ?urc^* \<and> ((?ts i, ?ss (Suc i)) \<in> ?ur^* \<or> (?tIsig i, ?sIsig (Suc i)) \<in> ?R)" (is "\<forall> i. ?goal1 i \<and> ?goal2 i")
  proof
    fix i
    from steps snt have snti: "SN_elt (rstep (set R)) (?ts i)" and stepst: "(?ts i, ?ss (Suc i)) \<in> (rstep (set R))^*" by auto
    from this urClosed have one: "(?I (?ts i), ?I (?ss (Suc i))) \<in> ?urc^*
      \<and> ((?I (?ts i), ?I (?ss (Suc i))) \<in> ?R \<or> (?ts i, ?ss (Suc i)) \<in> ?ur^*)" (is "_ \<and> (?IR \<or> ?oU)")
      by (rule iTransVI)
    
    from allP have "(s i, t i) \<in> P" by auto 
    with urPClosed have "urClosedT (t i) R ur" by auto
    from this snti have It: "?I (?ts i) = t i \<cdot> ?Is i"  by (rule iTransI)
    with one have two: "(?tIsig i, ?I (?ss (Suc i))) \<in> ?urc^*" by auto
    
    have "?ce^* \<subseteq> ?urc^*" by (rule rtrancl_mono, rule rstep_mono, auto) 
    from stepst snti have "SN_elt (rstep (set R)) (?ss (Suc i))" by (rule rtrancl_step_preserves_SN[where R = "rstep(set R)"])
    hence "(?I (?ss (Suc i)), s (Suc i) \<cdot> ?Is (Suc i)) \<in> ?ce^*" by (rule iTransII)
    with `?ce^* \<subseteq> ?urc^*` have Is: "(?I (?ss (Suc i)), ?sIsig (Suc i)) \<in> ?urc^*" by auto
    
    with two have part1: "?goal1 i" by auto
    
    from one have "?IR \<or> ?oU" ..
    hence "?goal2 i"
    proof
      assume "?oU" thus ?thesis ..
    next
      assume ?IR
      with It Is have "(?tIsig i, ?sIsig (Suc i)) \<in> ?R" by force
      thus ?thesis ..
    qed
    with part1 
    show "?goal1 i \<and> ?goal2 i" ..
  qed

  with chain allP snt show ?thesis by auto
qed

lemma redPairUrSound :
  assumes rules: "ur \<subseteq> NS"
  and nonStrict: "rP \<inter> P \<subseteq> NS \<union> S"
  and strict: "P - rP \<subseteq> S"
  and urClosed: "urClosed ur R"
  and urPClosed: "urPClosed R ur P"
  and minChain: "min_ichain (P,set R) s t \<sigma>"
  shows  "\<exists> j. min_ichain (rP, set R) (shift s j) (shift t j) (shift \<sigma> j)"
proof -
  let ?urc = "ur \<union> ceTrs (c,d,n)"
  from rules NS_ceCompat have urc: "?urc \<subseteq> NS" by auto
  from minChain urClosed urPClosed have "ichain (P, ?urc) s t (\<lambda> i. iTransSig (R,ur,(c,d,n)) (\<sigma> i))" 
    (is "ichain _ _ _ ?Isig") using iTransSound by blast
  from urc nonStrict strict this have "\<exists>j. ichain (rP, ?urc) (shift s j) (shift t j) (shift ?Isig j)" by (rule redPairSound)
  from this obtain j where ichain: "ichain (rP, ?urc) (shift s j) (shift t j) (shift ?Isig j)" ..
  have "min_ichain (rP, set R) (shift s j) (shift t j) (shift \<sigma> j)" 
    using min_ichainI[where R' = "set R" and R = "set R"] using ichain minChain ci_subset_refl[of "set R"] by auto
  thus ?thesis ..
qed
end

locale mono_ce_redpair = ce_redpair + 
  assumes S_ceCompat: "ceTrs (c,d,n) \<subseteq> \<succ>"
begin
lemma monoRedPairUrSound :
  assumes monotonic: "one_step_ctxt_closed S"
  and rNonStrict: "rR \<inter> ur \<subseteq> NS \<union> S"
  and rStrict: "ur - rR \<subseteq> S"
  and nonStrict: "rP \<inter> P \<subseteq> NS \<union> S"
  and strict: "P - rP \<subseteq> S"
  and subset: "rR \<subseteq> set R"
  and urClosed: "urClosed ur R"
  and urPClosed: "urPClosed R ur P"
  and minChain: "min_ichain (P,set R) s t \<sigma>"
  shows  "\<exists> j. min_ichain (rP, rR) (shift s j) (shift t j) (shift \<sigma> j)"
proof -
  let ?urc = "ur \<union> ceTrs (c,d,n)"
  let ?rurc = "rstep ?urc"
  let ?Isig = "(\<lambda> i. iTransSig (R,ur,(c,d,n)) (\<sigma> i))"
  let ?c = "rstep (ceTrs (c,d,n))"
  obtain \<delta> where delta: "\<delta> = ?Isig" by auto
  from rStrict rNonStrict NS_ceCompat have urc: "?urc \<subseteq> NS \<union> S" by auto
  from minChain urClosed urPClosed have ichain: "ichain (P, ?urc) s t \<delta> \<and>
  (\<forall> i. (t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (rstep ur)^* \<or> 
         (t i \<cdot> \<delta> i,
          s (Suc i) \<cdot> \<delta> (Suc i)) \<in> 
          ?rurc^* O ?c O ?rurc^*)"
    (is "?one \<and> (\<forall> i. ?cas1 i \<or> ?cas2 i)") by (simp only: delta, rule  iTransSound) 
  hence "?one" ..
  hence "\<exists> j. ichain (P, ur) (shift s j) (shift t j) (shift \<sigma> j)"
  proof (cases "\<exists> j. \<forall> k. ?cas1 (k + j)")
    case True
    from this obtain j where "\<forall> k. ?cas1 (k + j)" by auto
    with minChain have "ichain (P, ur) (shift s j) (shift t j) (shift \<sigma> j)" by auto
    thus ?thesis ..
  next
    case False
    with ichain have infCe: "\<forall> j. \<exists> k. ?cas2 (k + j)" by blast
    from subS monotonic S_ceCompat trs_inclusion have orientCe: "rstep (ceTrs (c,d,n)) \<subseteq> \<succ>" by blast
    have "\<forall> i. (s i \<cdot> \<delta> i, t i \<cdot> \<delta> i) \<in> \<succeq> \<union> \<succ>" (is "\<forall> i. _ \<in> ?NSS")
    proof 
      fix i
      show "(s i \<cdot> \<delta> i, t i \<cdot> \<delta> i) \<in> ?NSS"
      proof (cases "(s i, t i) \<in> \<succ>")
	case True
	with subS show ?thesis unfolding sub_closed_def using csI by blast
      next
	case False
	with strict nonStrict minChain have "(s i, t i) \<in> \<succeq>" by auto
	with subNS show ?thesis unfolding sub_closed_def using csI by blast
      qed
    qed
    hence orientP: "\<forall> i. (s i \<cdot> \<delta> i, t i \<cdot> \<delta> i) \<in> ?NSS^*" by auto

    have orientUrc: "rstep ?urc \<subseteq> ?NSS"  
    proof 
      fix st
      assume st: "st \<in> rstep ?urc"
      show "st \<in> ?NSS"
      proof (cases st)
	case (Pair s t)
	with st have "(s,t) \<in> rstep ?urc" by simp
	hence "(s,t) \<in> ?NSS"
	proof
	  fix ls rs C
	  assume s: "s = C\<langle>ls\<rangle>" and t: "t = C\<langle>rs\<rangle>" and CS: "(ls,rs) \<in> CS ?urc"
	  from CS show ?thesis
	  proof
	    fix l r \<sigma>
	    assume l: "ls = l \<cdot> \<sigma>" and r: "rs = r \<cdot> \<sigma>" and lr: "(l,r) \<in> ur \<union> ceTrs (c,d,n)"
	    with S_ceCompat rStrict rNonStrict have "(l,r) \<in> \<succeq> \<union> \<succ>" by blast
	    thus ?thesis
	    proof
	      assume "(l,r) \<in> \<succeq>" with l r subNS have "(ls,rs) \<in> \<succeq>" unfolding sub_closed_def using csI by auto
	      with s t wmNS one_step_ctxt_closed_imp_ctxt_closed show ?thesis unfolding ctxt_closed_def using ccI by auto
	    next
	      assume "(l,r) \<in> \<succ>" with l r subS have "(ls,rs) \<in> \<succ>" unfolding sub_closed_def using csI by auto
	      with s t monotonic one_step_ctxt_closed_imp_ctxt_closed show ?thesis unfolding ctxt_closed_def using ccI by auto
	    qed
	  qed
	qed
	with Pair show ?thesis by auto
      qed
    qed
	
    from orientUrc have orientUrcs: "(rstep ?urc)^* \<subseteq> ?NSS^*" by (rule rtrancl_mono)
    with `?one` have orientSteps: "\<forall> i. (t i \<cdot> \<delta> i, s (Suc i) \<cdot> \<delta> (Suc i)) \<in> ?NSS^*" by auto

    from orientUrcs orientCe infCe have 
      orientStrict: "\<forall> j. \<exists> k. (t (k + j) \<cdot> \<delta> (k+j), s (Suc (k + j)) \<cdot> \<delta> (Suc (k+j))) \<in> ?NSS^* O \<succ> O ?NSS^*" by blast
    
    let ?S = "?NSS^* O \<succ> O ?NSS^*"
    let ?NS = "?NSS^*"
    let ?f = "\<lambda> i. (t i \<cdot> \<delta> i)"
    obtain f where f: "f = ?f" by auto
    have nonStrict: "\<forall> i. (f i, f (Suc i)) \<in> ?NS \<union> ?S" using orientP orientSteps trans_rtrancl[where r = ?NSS] unfolding trans_def 
      by (simp only: f, blast)
    have compat1: "?S O ?NS \<subseteq> ?S" using rtrancl_trans by force
    have compat2: "?NS O ?S \<subseteq> ?S" using rtrancl_trans by force
    have infStrict: "\<forall> j. \<exists> k \<ge> j. (f k, f (Suc k)) \<in> ?S"      
    proof 
      fix j
      from orientStrict have "\<exists> k. (f (k + j), s (Suc (k+j)) \<cdot> \<delta> (Suc (k + j))) \<in> ?S" by (simp only: f)
      from this obtain k where "(f (k + j), s (Suc (k + j)) \<cdot> \<delta> (Suc (k + j))) \<in> ?S" (is "(_,?skj) \<in> _") by auto
      with orientP have "(f (k+j), f (Suc (k+j))) \<in> ?S O ?NS" by (auto simp: f)
      with compat1 have "k + j \<ge> j" and "(f (k+j), f (Suc (k+j))) \<in> ?S" by (auto simp: O_assoc)
      thus "\<exists> k \<ge> j. (f k, f (Suc k)) \<in> ?S" by blast
    qed
    from rtrancl_S_rtrancl_is_SN[where NS = NS and S = S] SN compat 
      have "SN_elt ?S (f 0)" unfolding SN_def by force
    from nonStrict compat2 this have "\<exists> j. \<forall> i \<ge> j. (f i, f (Suc i)) \<in> ?NS - ?S" by (rule non_strict_ending)
    with infStrict have False by blast    
    thus ?thesis ..
  qed
  from this obtain j where "ichain (P,ur) (shift s j) (shift t j) (shift \<sigma> j)" ..
  
  with monotonic rNonStrict rStrict nonStrict strict 
  have "\<exists> k. ichain (rP,rR) (shift (shift s j) k) (shift (shift t j) k) (shift (shift \<sigma> j) k)" (is "\<exists> k. ?chain k")
    by (rule monoRedPairSound)
  from this obtain k where kjChain: "?chain k" ..
  have shift: "\<And> f. shift (shift f j) k = shift f (k + j)" 
  proof (rule ext, simp)
    fix f x
    have "x + k + j = (x + ( k + j))" by auto
    thus "f (x + k + j) = f (x + ( k + j))" by (rule arg_cong)
  qed
  from kjChain have inf_chain: "ichain (rP,rR) (shift s (k + j)) (shift t (k + j)) (shift \<sigma> (k + j))" 
    by (simp only: shift[where f = s] shift[where f = t] shift[where f = \<sigma>])
  have "min_ichain (rP, rR) (shift s (k + j)) (shift t (k + j)) (shift \<sigma> (k + j))" 
    using min_ichainI[where R' = "set R" and R = "rR"] using inf_chain minChain ci_subsetI[OF subset]  by auto
  thus ?thesis ..
qed
end

abbreviation redPairUrProc :: "('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> bool"
 where "redPairUrProc S NS ur P R rm rrm rP \<equiv> urClosed ur R \<and> urPClosed R ur (set P) \<and> ur \<subseteq> NS \<and> (set rP) \<inter> (set P) \<subseteq> NS \<union> S \<and> (set P) - (set rP) \<subseteq> S"

abbreviation monoRedPairUrProc :: "('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> bool"
 where "monoRedPairUrProc S NS ur P R rm rP rR \<equiv> one_step_ctxt_closed S \<and> urClosed ur R \<and> urPClosed R ur (set P) \<and> (set rR) \<inter> ur \<subseteq> NS \<union> S \<and> ur - (set rR) \<subseteq> S \<and> (set rP) \<inter> (set P) \<subseteq> NS \<union> S \<and> (set P) - (set rP) \<subseteq> S \<and> set rR \<subseteq> set R"

context ce_redpair
begin
lemma redPairUrProc_ci : "chain_identifying_onlyDP_proc_impl (redPairUrProc \<succ> \<succeq> ur)"
unfolding chain_identifying_onlyDP_proc_impl_def
proof (intro allI, intro impI)
  fix P R rP 
  assume proc: "redPairUrProc \<succ> \<succeq> ur P R rm rrm rP"
  show "\<forall> s t \<sigma> . min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i))"
  proof (intro allI, intro impI)
    fix s t \<sigma>
    assume min: "min_ichain (set P, set R) s t \<sigma>"
    with proc have "\<exists> j. min_ichain (set rP, set R) (shift s j) (shift t j) (shift \<sigma> j)" using redPairUrSound by auto
    thus "\<exists> j. ichain (set rP, set R) (shift s j) (shift t j) (shift \<sigma> j)" by force
  qed
qed
end

context mono_ce_redpair
begin
lemma monoRedPairUrProc_ci : "chain_identifying_proc_impl (monoRedPairUrProc \<succ> \<succeq> ur)"
unfolding chain_identifying_proc_impl_def
proof (intro allI, intro impI)
  fix P R rP rR
  assume proc: "monoRedPairUrProc \<succ> \<succeq> ur P R rm rP rR"
  show "set rR \<subseteq>ci set R \<and> (\<forall> s t \<sigma> . min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set rR) (shift s i) (shift t i) (shift \<sigma> i)))"
  proof (intro conjI, simp add: proc ci_subsetI[of "set rR" "set R"], intro allI, intro impI)
    fix s t \<sigma>
    assume min: "min_ichain (set P, set R) s t \<sigma>"
    with proc have "\<exists> j. min_ichain (set rP, set rR) (shift s j) (shift t j) (shift \<sigma> j)" using monoRedPairUrSound by auto
    thus "\<exists> j. ichain (set rP, set rR) (shift s j) (shift t j) (shift \<sigma> j)" by force
  qed
qed
end

context ce_af_redpair 
begin
lemma redPairUrAfSound :
  assumes rules: "ur \<subseteq> NS"
  and nonStrict: "rP \<inter> P \<subseteq> NS \<union> S"
  and strict: "P - rP \<subseteq> S"
  and urClosed: "urClosed_af ur R \<pi>"
  and urPClosed: "urPClosed_af R ur P \<pi>"
  and minChain: "min_ichain (P,set R) s t \<sigma>"
  shows  "\<exists> j. min_ichain (rP, set R) (shift s j) (shift t j) (shift \<sigma> j)"
proof -
  let ?ss = "\<lambda> i. s i \<cdot> \<sigma> i"
  let ?ts = "\<lambda> i. t i \<cdot> \<sigma> i"
  let ?I = "iTrans (R, ur, (c,d,n))"
  let ?Is = "\<lambda> i. (iTransSig (R,ur,(c,d,n)) (\<sigma> i))"
  let ?sIsig = "\<lambda> i. (s i) \<cdot> ?Is i"
  let ?tIsig = "\<lambda> i. (t i) \<cdot> ?Is i"
  from minChain have snt: "\<forall> i. SN_elt (rstep (set R)) (?ts i)"
    and steps: "\<forall> i. (?ts i, ?ss (Suc i)) \<in> (rstep (set R))^*"
    and P: "\<forall> i. (s i, t i) \<in> P"
    by (auto simp: minimal_cond_def)
  from steps snt have sns: "\<forall> i. SN_elt (rstep (set R)) (?ss (Suc i))" using rtrancl_step_preserves_SN[where R = "rstep (set R)"] by blast
  from P urPClosed have urClosedT: "\<forall> i. urClosedT_af (t i) R ur \<pi>" by force
  have nsSteps: "\<forall> i. (?tIsig i, ?sIsig (Suc i)) \<in> NS^*"
  proof
    fix i
    have "(?tIsig i, ?I (?ts i)) \<in> NS^*" by (rule iTransI_af, auto simp: urClosedT snt)
    also have "(?I (?ts i), ?I (?ss (Suc i))) \<in> NS^*" by (rule iTransVI_af, auto simp: snt urClosed rules steps)
    also have "(?I (?ss (Suc i)), ?sIsig (Suc i)) \<in> NS^*" by (rule iTransII_af, auto simp: sns)
    finally show "(?tIsig i, ?sIsig (Suc i)) \<in> NS^*" .
  qed
  have "NS = (rstep NS)" 
  proof
    show "NS \<subseteq> rstep NS" using rstep_contains_r .
  next
    show "rstep NS \<subseteq> NS" by (rule trs_inclusion, auto simp: wmNS subNS)
  qed
  with P nsSteps have ichain: "ichain (P, NS) s t ?Is" by auto
  
  have "\<exists> j. ichain (rP, NS) (shift s j) (shift t j) (shift ?Is j)"
    by (rule redPairSound, simp, rule nonStrict, rule strict, rule ichain)
  from this obtain j where "ichain (rP, NS) (shift s j) (shift t j) (shift ?Is j)" ..
  with minChain have ichain: "ichain (rP, set R) (shift s j) (shift t j) (shift \<sigma> j)" by auto
  have "min_ichain (rP, set R) (shift s j) (shift t j) (shift \<sigma> j)" 
    using min_ichainI[where R' = "set R" and R = "set R"] using ichain minChain ci_subset_refl[of "set R"] by auto
  thus ?thesis ..
qed
end

abbreviation redPairUrAfProc :: "('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> 'f af \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> bool"
 where "redPairUrAfProc S NS ur \<pi> P R rm rrm rP \<equiv> urClosed_af ur R \<pi> \<and> urPClosed_af R ur (set P) \<pi> \<and> ur \<subseteq> NS \<and> (set rP) \<inter> (set P) \<subseteq> NS \<union> S \<and> (set P) - (set rP) \<subseteq> S"

context ce_af_redpair 
begin

lemma redPairUrAfProc_ci : "chain_identifying_onlyDP_proc_impl (redPairUrAfProc \<succ> \<succeq> ur \<pi>)"
unfolding chain_identifying_onlyDP_proc_impl_def
proof (intro allI, intro impI)
  fix P R rP 
  assume proc: "redPairUrAfProc \<succ> \<succeq> ur \<pi> P R rm rrm rP"
  show "\<forall> s t \<sigma> . min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i))"
  proof (intro allI, intro impI)
    fix s t \<sigma>
    assume min: "min_ichain (set P, set R) s t \<sigma>"
    with proc have "\<exists> j. min_ichain (set rP, set R) (shift s j) (shift t j) (shift \<sigma> j)" using redPairUrAfSound by auto
    thus "\<exists> j. ichain (set rP, set R) (shift s j) (shift t j) (shift \<sigma> j)" by force
  qed
qed
end


fun matchCapBelowI :: "('f,'v)term \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)term \<Rightarrow> bool"
  where "matchCapBelowI l R (Fun f ts) = gcMatchI (HFun f (map (capI R) ts)) l"

lemma matchCapBelowISound[simp] :  "matchCapBelowI l R (Fun f ts) = matchCapBelow l R (Fun f ts)"
proof - 
  have cap: "capI R = cap (set R)"
    by (rule ext, simp add: capISound)
  have gcMatch: "gcMatchI = gcMatch"
    by (rule ext, rule ext, simp add: gcMatchISound)
  show ?thesis by (simp only: cap gcMatch matchCapBelowI.simps matchCapBelow.simps)
qed

declare matchCapBelowI.simps[simp del]

fun matchCapRMBelow :: "('f,'v)term \<Rightarrow> ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)term \<Rightarrow> bool"
  where "matchCapRMBelow l rm (Fun f ts) = gcMatchI (HFun f (map (capRM rm) ts)) l"

lemma matchCapRMBelowSound[simp] :  "matchCapRMBelow l (computeRuleMap R) (Fun f ts) = matchCapBelow l R (Fun f ts)"
proof - 
  have cap: "capRM (computeRuleMap R) = cap (set R)"
    by (rule ext, simp add: capRMSound)
  have gcMatch: "gcMatchI = gcMatch"
    by (rule ext, rule ext, simp add: gcMatchISound)
  show ?thesis by (simp only: cap gcMatch matchCapRMBelow.simps matchCapBelow.simps)
qed

declare matchCapRMBelow.simps[simp del]

fun checkUrClosedT :: "('f::shows,'v::shows)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)term \<Rightarrow> shows monad"
where "checkUrClosedT _ _ (Var x) = OK" 
   |  "checkUrClosedT R ur (Fun f ts) = (
  checkAllM (checkUrClosedT R ur) ts
 \<guillemotright> checkAllM (\<lambda> lr. checkB (lr mem ur \<or> \<not> matchCapBelowI (fst lr) R (Fun f ts))
   (''due to the subterm '' +#+ shows(Fun f ts) +@+ '' of some usable rhs, rule '' +#+ shows_rule lr +@+ shows_string '' should be usable.'')
   ) R
)"

fun checkUrClosedT_af :: "('f::shows,'v::shows)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 'f af \<Rightarrow> ('f,'v)term \<Rightarrow> shows monad"
where "checkUrClosedT_af _ _ _ (Var x) = OK" 
   |  "checkUrClosedT_af R ur \<pi> (Fun f ts) = (let n = length ts in (
   checkAllM (\<lambda>i. if \<pi> f n i then checkUrClosedT_af R ur \<pi> (ts ! i) else OK) [0 ..< n]
 \<guillemotright> checkAllM (\<lambda>lr. checkB (lr mem ur \<or> \<not> matchCapBelowI (fst lr) R (Fun f ts)) 
     (''due to the subterm '' +#+ shows(Fun f ts) +@+ '' of some usable rhs, rule '' +#+ shows_rule lr +@+ shows_string '' should be usable.'')
   ) R
))"

fun checkUrClosedTRM :: "('f \<Rightarrow> ('f,'v)trsL) \<Rightarrow> ('f::shows,'v::shows)trsL \<Rightarrow> ('f,'v)term \<Rightarrow> shows monad"
where "checkUrClosedTRM _ _ (Var x) = OK" 
   |  "checkUrClosedTRM rm ur (Fun f ts) = (
   checkAllM (checkUrClosedTRM rm ur) ts
 \<guillemotright> checkAllM (\<lambda> lr. checkB (lr mem ur \<or> \<not> matchCapRMBelow (fst lr) (Some rm) (Fun f ts)) 
     (''due to the subterm '' +#+ shows(Fun f ts) +@+ '' of some usable rhs, rule '' +#+ shows_rule lr +@+ shows_string '' should be usable.'')
   ) (rm f)
)"

fun checkUrClosedTRM_af :: "('f \<Rightarrow> ('f,'v)trsL) \<Rightarrow> ('f::shows,'v::shows)trsL \<Rightarrow> 'f af \<Rightarrow> ('f,'v)term \<Rightarrow> shows monad"
where "checkUrClosedTRM_af _ _ _ (Var x) = OK" 
   |  "checkUrClosedTRM_af rm ur \<pi> (Fun f ts) = (let n = length ts in (
   checkAllM (\<lambda> i. if \<pi> f n i then checkUrClosedTRM_af rm ur \<pi> (ts ! i) else OK) [0 ..< n]
 \<guillemotright> checkAllM (\<lambda> lr. checkB (lr mem ur \<or> \<not> matchCapRMBelow (fst lr) (Some rm) (Fun f ts)) 
    (''due to the subterm '' +#+ shows(Fun f ts) +@+ '' of some usable rhs, rule '' +#+ shows_rule lr +@+ shows_string '' should be usable.'')
   ) (rm f)
))" 

lemma checkUrClosedTSound[simp]:
  assumes "isOK(checkUrClosedT R ur t)" 
  shows "urClosedT t R (set ur)"
using assms
proof (induct t, simp)
  case (Fun f ts)
  hence ind: "\<forall> t \<in> set ts. urClosedT t R (set ur)" and
    rls: "\<forall> (l,r) \<in> set R. (l,r) \<in> set ur \<or> \<not> matchCapBelow l R (Fun f ts)" by (auto simp: mem_iff)
  thus ?case by auto
qed

lemma checkUrClosedT_afSound[simp] : assumes "isOK(checkUrClosedT_af R ur \<pi> t)" 
  shows "urClosedT_af t R (set ur) \<pi>"
using assms
proof (induct t, simp)
  case (Fun f ts)
  hence ind: "\<forall> i < length ts. \<pi> f (length ts) i \<longrightarrow> urClosedT_af (ts ! i) R (set ur) \<pi>" and
    rls: "\<forall> (l,r) \<in> set R. (l,r) \<in> set ur \<or> \<not> matchCapBelow l R (Fun f ts)" by (auto simp: Let_def mem_iff) force
  thus ?case by auto
qed

lemma checkUrClosedTRMSound[simp] : assumes check: "isOK(checkUrClosedTRM rm ur t)" 
  and ruleMap: "computeRuleMap R = Some rm"
  shows "urClosedT t R (set ur)"
using assms
proof (induct t, simp)
  case (Fun f ts)
  hence ind: "\<forall> t \<in> set ts. urClosedT t R (set ur)" and
    rls: "\<forall> (l,r) \<in> set (rm f). (l,r) \<in> set ur \<or> \<not> matchCapRMBelow l (computeRuleMap R) (Fun f ts)" by (auto simp: mem_iff)
  from ruleMap computeRuleMapSound2[where R = R]  have noLeftVar: "\<forall> (l,r) \<in> set R. root l \<noteq> None" by auto
  from ruleMap computeRuleMapSound[where R = R and rm = rm and f = f] have fLookup: "set (rm f) = {(l,r) | l r. (l,r) \<in> set R \<and> root l = Some f}" by auto
  have "\<forall> (l,r) \<in> set R. (l,r) \<in> set ur \<or> \<not> matchCapRMBelow l (computeRuleMap R) (Fun f ts)" (is "\<forall> lr \<in> set R. ?P lr") 
  proof
    fix lr
    assume lr: "lr \<in> set R"
    show "?P lr"
    proof (cases lr)
      case (Pair l r)
      show ?thesis
      proof (cases "(l,r) \<in> set (rm f)")
	case True
	with Pair rls show ?thesis by auto
      next
	case False
	from noLeftVar lr Pair obtain g ss where l: "l = Fun g ss" by (cases l, auto)
	with fLookup False lr Pair have "f \<noteq> g" by auto
	with l ruleMap have "\<not> matchCapRMBelow l (computeRuleMap R) (Fun f ts)" by (auto simp: matchCapRMBelow.simps)
	with Pair rls show ?thesis by auto
      qed
    qed
  qed
  with ind show ?case by auto
qed

lemma checkUrClosedTRM_afSound[simp] : assumes check: "isOK(checkUrClosedTRM_af rm ur \<pi> t)" 
  and ruleMap: "computeRuleMap R = Some rm"
  shows "urClosedT_af t R (set ur) \<pi>"
using assms
proof (induct t, simp)
  case (Fun f ts)
  hence ind: "\<forall> i < length ts. \<pi> f (length ts) i \<longrightarrow> urClosedT_af (ts ! i) R (set ur) \<pi>" and
    rls: "\<forall> (l,r) \<in> set (rm f). (l,r) \<in> set ur \<or> \<not> matchCapRMBelow l (computeRuleMap R) (Fun f ts)" by (auto simp: Let_def mem_iff) force
  from ruleMap computeRuleMapSound2[where R = R]  have noLeftVar: "\<forall> (l,r) \<in> set R. root l \<noteq> None" by auto
  from ruleMap computeRuleMapSound[where R = R and rm = rm and f = f] have fLookup: "set (rm f) = {(l,r) | l r. (l,r) \<in> set R \<and> root l = Some f}" by auto
  have "\<forall> (l,r) \<in> set R. (l,r) \<in> set ur \<or> \<not> matchCapRMBelow l (computeRuleMap R) (Fun f ts)" (is "\<forall> lr \<in> set R. ?P lr") 
  proof
    fix lr
    assume lr: "lr \<in> set R"
    show "?P lr"
    proof (cases lr)
      case (Pair l r)
      show ?thesis
      proof (cases "(l,r) \<in> set (rm f)")
	case True
	with Pair rls show ?thesis by auto
      next
	case False
	from noLeftVar lr Pair obtain g ss where l: "l = Fun g ss" by (cases l, auto)
	with fLookup False lr Pair have "f \<noteq> g" by auto
	with l ruleMap have "\<not> matchCapRMBelow l (computeRuleMap R) (Fun f ts)" by (auto simp: matchCapRMBelow.simps)
	with Pair rls show ?thesis by auto
      qed
    qed
  qed
  with ind show ?case by auto
qed

fun checkUrPClosed :: "('f::shows,'v::shows)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> shows monad"
where "checkUrPClosed R ur P = (
    checkAllM (\<lambda> lr. checkUrClosedT R ur (snd lr)) ur
      \<leftarrow> (\<lambda>s. ''error when checking closure properties of rhs of usable rules'' +#+ shows_nl +@+ s)
  \<guillemotright> checkAllM (\<lambda> st. checkUrClosedT R ur (snd st)) P
      \<leftarrow> (\<lambda>s. ''error when checking closure properties of rhs of DPs'' +#+ shows_nl +@+ s)
  )"

fun checkUrPClosed_af :: "('f::shows,'v::shows)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 'f af \<Rightarrow> shows monad"
where "checkUrPClosed_af R ur P \<pi> = (
    checkAllM (\<lambda> lr. checkUrClosedT_af R ur \<pi> (snd lr)) ur
      \<leftarrow> (\<lambda>s. ''error when checking closure properties of rhs of usable rules'' +#+ shows_nl +@+ s)
  \<guillemotright> checkAllM (\<lambda> st. checkUrClosedT_af R ur \<pi> (snd st)) P
      \<leftarrow> (\<lambda>s. ''error when checking closure properties of rhs of DPs'' +#+ shows_nl +@+ s)
  )"

lemma checkUrPClosedSound[simp] : assumes "isOK(checkUrPClosed R ur P)"
  shows "urClosed (set ur) R" and "urPClosed R (set ur) (set P)"
using assms by auto

declare checkUrPClosed.simps[simp del]


lemma checkUrPClosedSound_af[simp] : assumes "isOK(checkUrPClosed_af R ur P \<pi>)"
  shows "urClosed_af (set ur) R \<pi>" and "urPClosed_af R (set ur) (set P) \<pi>"
using assms by auto

declare checkUrPClosed_af.simps[simp del]


fun checkUrPClosedRM :: "('f::shows,'v::shows)trsL \<Rightarrow> ('f \<Rightarrow> ('f,'v)trsL) option \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> shows monad"
where "checkUrPClosedRM R rmo ur P = (case rmo of None \<Rightarrow> checkUrPClosed R ur P | Some rm \<Rightarrow> (
    checkAllM (\<lambda> lr. checkUrClosedTRM rm ur (snd lr)) ur
      \<leftarrow> (\<lambda>s. ''error when checking closure properties of rhs of usable rules'' +#+ shows_nl +@+ s)
  \<guillemotright> checkAllM (\<lambda> st. checkUrClosedTRM rm ur (snd st)) P
      \<leftarrow> (\<lambda>s. ''error when checking closure properties of rhs of DPs'' +#+ shows_nl +@+ s)
  ))"

fun checkUrPClosedRM_af :: "('f::shows,'v::shows)trsL \<Rightarrow> ('f \<Rightarrow> ('f,'v)trsL) option \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 'f af \<Rightarrow> shows monad"
where "checkUrPClosedRM_af R rmo ur P \<pi> = (case rmo of None \<Rightarrow> checkUrPClosed_af R ur P \<pi> | Some rm \<Rightarrow> (
    checkAllM (\<lambda> lr. checkUrClosedTRM_af rm ur \<pi> (snd lr)) ur
      \<leftarrow> (\<lambda>s. ''error when checking closure properties of rhs of usable rules'' +#+ shows_nl +@+ s)
  \<guillemotright> checkAllM (\<lambda> st. checkUrClosedTRM_af rm ur \<pi> (snd st)) P
      \<leftarrow> (\<lambda>s. ''error when checking closure properties of rhs of DPs'' +#+ shows_nl +@+ s)
  ))"

lemma checkUrPClosedRMSound[simp] : assumes "isOK(checkUrPClosedRM R (computeRuleMap R) ur P)"
  shows "urClosed (set ur) R \<and> urPClosed R (set ur) (set P)"
using assms
by (cases "computeRuleMap R", auto)

declare checkUrPClosedRM.simps[simp del]


lemma checkUrPClosedSoundRM_af[simp] : assumes "isOK(checkUrPClosedRM_af R (computeRuleMap R) ur P \<pi>)"
  shows "urClosed_af (set ur) R \<pi> \<and> urPClosed_af R (set ur) (set P) \<pi>"
using assms by (cases "computeRuleMap R", auto)

declare checkUrPClosedRM_af.simps[simp del]

fun checkRedPairUrProc :: "
  (('f::shows,'v::shows)rule \<Rightarrow> shows monad) \<Rightarrow>
  (('f,'v)rule \<Rightarrow> shows monad) \<Rightarrow> 
   shows \<Rightarrow>
   shows \<Rightarrow> 
   ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> shows monad"
where "checkRedPairUrProc checkSS checkNSS  name representation P rP r rm ur  = (
  let (Sp :: ('f,'v) trsL ,NSp) = List.partition (\<lambda> foo. isOK(checkSS foo)) P in (
   checkUrPClosedRM r rm ur P
    \<leftarrow> (\<lambda>s. ''problem with closure properties of usable rules'' +#+ shows_nl +@+ s) 
 \<guillemotright> checkAllM checkNSS ur
    \<leftarrow> (\<lambda>s. ''problem when orienting usable rules'' +#+ shows_nl +@+ s) 
 \<guillemotright> checkAllM checkNSS NSp
    \<leftarrow> (\<lambda>s. ''problem when orienting DPs'' +#+ shows_nl +@+ s)
 \<guillemotright> checkSubset NSp rP
    \<leftarrow> (\<lambda>s. case checkSS s of Left err \<Rightarrow> err | Right ok \<Rightarrow> internal_error)
)  \<leftarrow> (\<lambda>s. ''could not apply the reduction pair processor with the following'' +#+ shows_nl +@+
          representation +@+ shows_nl +@+ ''and the following usable rules'' +#+ shows_nl +@+ shows_rules ur +@+ shows_nl +@+ s)
)"

locale ce_redpair_impl = redpair_impl + ce_redpair
begin

lemma check_redpair_ur_proc_sound[simp] :
  shows "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(checkRedPairUrProc checkS checkNS name repr P rP R rm ur))"
unfolding chain_identifying_onlyDP_proc_impl_def
proof (intro allI, intro impI)
  fix P R rP 
  assume check: "isOK (checkRedPairUrProc checkS checkNS name repr P rP R (computeRuleMap R) ur)"
  have "redPairUrProc \<succ> \<succeq> (set ur) P R (computeRuleMap R) rrm rP" 
  proof (cases "List.partition (\<lambda> dp. isOK(checkS dp)) P")
    case (Pair Sp NSp)
    let ?NSp = "set NSp \<inter> set rP"
    from Pair check checkSSound checkNSSound have cond:
      "set P - set NSp \<subseteq> S \<and> 
      ?NSp \<inter> set P \<subseteq> NS \<union> S   \<and> 
      set ur \<subseteq> NS     \<and>
      set NSp \<subseteq> set rP  \<and>
      isOK(checkUrPClosedRM R (computeRuleMap R) ur P)"
      by (auto simp: Let_def mem_iff)
    hence id: "set NSp = ?NSp" and stri: "set P - ?NSp \<subseteq> S" by (auto simp: mem_iff)
    from cond have "isOK(checkUrPClosedRM R (computeRuleMap R) ur P)" by auto
    hence "urClosed (set ur) R \<and> urPClosed R (set ur) (set P)" using checkUrPClosedRMSound by auto
    with cond id stri show ?thesis by auto
  qed
  with redPairUrProc_ci show "\<forall> s t \<sigma>. min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i))"
    unfolding chain_identifying_onlyDP_proc_impl_def by auto
qed
end

fun checkRedPairUrAfProc :: "
  (('f::shows,'v::shows)rule \<Rightarrow> shows monad) \<Rightarrow>
  (('f,'v)rule \<Rightarrow> shows monad) \<Rightarrow> 
   shows \<Rightarrow>
   shows \<Rightarrow> 
   ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> 'f af \<Rightarrow> shows monad"
where "checkRedPairUrAfProc checkSS checkNSS  name representation P rP r rm ur \<pi> = (
  let (Sp :: ('f,'v) trsL ,NSp) = List.partition (\<lambda> foo. isOK(checkSS foo)) P in (
   checkUrPClosedRM_af r rm ur P \<pi>
    \<leftarrow> (\<lambda>s. ''problem with closure properties of usable rules w.r.t. argument filter'' +#+ shows_nl +@+ s) 
 \<guillemotright> checkAllM checkNSS ur
    \<leftarrow> (\<lambda>s. ''problem when orienting usable rules'' +#+ shows_nl +@+ s) 
 \<guillemotright> checkAllM checkNSS NSp
    \<leftarrow> (\<lambda>s. ''problem when orienting DPs'' +#+ shows_nl +@+ s)
 \<guillemotright> checkSubset NSp rP
    \<leftarrow> (\<lambda>s. case checkSS s of Left err \<Rightarrow> err | Right ok \<Rightarrow> internal_error
) \<leftarrow> (\<lambda>s. ''could not apply the reduction pair processor with the following'' +#+ shows_nl +@+
         representation +@+ shows_nl +@+ ''and the following usable rules'' +#+ shows_nl +@+ shows_rules ur +@+ shows_nl +@+ s)
))"

locale ce_af_redpair_impl = redpair_impl + ce_af_redpair

sublocale ce_af_redpair_impl \<subseteq> ce_redpair_impl ..

context ce_af_redpair_impl
begin

lemma check_redpair_ur_af_proc_sound[simp] :
  shows "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(checkRedPairUrAfProc checkS checkNS name repr P rP R rm ur \<pi>))"
unfolding chain_identifying_onlyDP_proc_impl_def
proof (intro allI, intro impI)
  fix P R rP 
  assume check: "isOK (checkRedPairUrAfProc checkS checkNS name repr P rP R (computeRuleMap R) ur \<pi>)"
  have "redPairUrAfProc \<succ> \<succeq> (set ur) \<pi> P R (computeRuleMap R) rrm rP" 
  proof (cases "List.partition (\<lambda> dp. isOK(checkS dp)) P")
    case (Pair Sp NSp)
    let ?NSp = "set NSp \<inter> set rP"
    from Pair check checkSSound checkNSSound have cond:
      "set P - set NSp \<subseteq> S \<and> 
      ?NSp \<inter> set P \<subseteq> NS \<union> S   \<and> 
      set ur \<subseteq> NS     \<and>
      set NSp \<subseteq> set rP  \<and>
      isOK(checkUrPClosedRM_af R (computeRuleMap R) ur P \<pi>)"
      by (auto simp: Let_def mem_iff)
    hence id: "set NSp = ?NSp" and stri: "set P - ?NSp \<subseteq> S" by (auto simp: mem_iff)
    from cond have "isOK(checkUrPClosedRM_af R (computeRuleMap R) ur P \<pi>)" by auto
    hence "urClosed_af (set ur) R \<pi> \<and> urPClosed_af R (set ur) (set P) \<pi>" using checkUrPClosedSoundRM_af by auto
    with cond id stri show ?thesis by auto
  qed
  with redPairUrAfProc_ci show "\<forall> s t \<sigma>. min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i))"
    unfolding chain_identifying_onlyDP_proc_impl_def by auto
qed
end


fun checkMonoRedPairUrProc :: "
   shows monad \<Rightarrow> 
  (('f::shows,'v::shows)rule \<Rightarrow> shows monad) \<Rightarrow>
  (('f,'v)rule \<Rightarrow> shows monad) \<Rightarrow> 
   shows \<Rightarrow>
   shows \<Rightarrow> 
   ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> shows monad"
where "checkMonoRedPairUrProc checkMono checkSS checkNSS  name representation P rP R rm rR ur = (
  let (Sp :: ('f,'v) trsL , NSp) = List.partition (\<lambda> foo. isOK(checkSS foo)) P in
  let (Sr :: ('f,'v) trsL , NSr) = List.partition (\<lambda> foo. isOK(checkSS foo)) ur in (
   checkMono
    \<leftarrow> (\<lambda>s. ''problem with monotonicity of strict order'' +#+ shows_nl +@+ s)
 \<guillemotright> checkUrPClosedRM R rm ur P
    \<leftarrow> (\<lambda>s. ''problem with closure properties of usable rules'' +#+ shows_nl +@+ s) 
 \<guillemotright> checkAllM checkNSS NSr
    \<leftarrow> (\<lambda>s. ''problem when orienting usable rules'' +#+ shows_nl +@+ s) 
 \<guillemotright> checkAllM checkNSS NSp
    \<leftarrow> (\<lambda>s. ''problem when orienting DPs'' +#+ shows_nl +@+ s)
 \<guillemotright> checkSubset rR R
    \<leftarrow> (\<lambda>s. ''problem: additional rule(s) to TRS added'' +#+ shows_nl)
 \<guillemotright> checkSubset NSr rR
    \<leftarrow> (\<lambda>s. case checkSS s of Left err \<Rightarrow> err | Right ok \<Rightarrow> internal_error)
 \<guillemotright> checkSubset NSp rP
    \<leftarrow> (\<lambda>s. case checkSS s of Left err \<Rightarrow> err | Right ok \<Rightarrow> internal_error)
) \<leftarrow> (\<lambda>s. ''could not apply the (monotonic) reduction pair processor with the following'' +#+ shows_nl +@+
         representation +@+ shows_nl +@+ ''and the following usable rules'' +#+ shows_rules ur +@+ shows_nl +@+ s)
)"

locale ce_mono_redpair_impl = mono_ce_redpair + mono_redpair_impl
begin

lemma check_mono_redpair_ur_proc_sound[simp] :
  shows "chain_identifying_proc_impl (\<lambda> P R rm rP rR. isOK(checkMonoRedPairUrProc checkMono checkS checkNS name repr P rP R rm rR ur))"
unfolding chain_identifying_proc_impl_def
proof (intro allI, intro impI)
  fix P R rP rR
  assume check: "isOK (checkMonoRedPairUrProc checkMono checkS checkNS name repr P rP R (computeRuleMap R) rR ur)"
  have mono: "monoRedPairUrProc \<succ> \<succeq> (set ur) P R (computeRuleMap R) rP rR" 
  proof (cases "List.partition (\<lambda> dp. isOK(checkS dp)) P")
    case (Pair Sp NSp)
    hence splitP:"List.partition (\<lambda> dp. isOK(checkS dp)) P = (Pair Sp NSp)" .
    let ?NSp = "set NSp \<inter> set rP"
    from Pair check checkSSound checkNSSound have cond:
      "one_step_ctxt_closed S \<and>
      set P - set NSp \<subseteq> S \<and> 
      ?NSp \<inter> set P \<subseteq> NS \<union> S   \<and> 
      set NSp \<subseteq> set rP  \<and>
      isOK(checkUrPClosedRM R (computeRuleMap R) ur P)  \<and>
      set rR \<subseteq> set R"
      by (auto simp: checkMonoSound Let_def mem_iff)    
    hence id: "set NSp = ?NSp" and stri: "set P - ?NSp \<subseteq> S" by (auto simp: mem_iff)
    show ?thesis
    proof (cases "List.partition (\<lambda> r. isOK(checkS r)) ur")
      case (Pair Sr NSr)
      let ?NSr = "set NSr \<inter> set rR"
      from Pair check checkSSound checkNSSound splitP have condR:
      "set ur - set NSr \<subseteq> S \<and> 
      ?NSr \<inter> set ur \<subseteq> NS \<union> S   \<and> 
      set NSr \<subseteq> set rR"
	by (auto simp: Let_def mem_iff)    
      hence idr: "set NSr = ?NSr" and strir: "set ur - ?NSr \<subseteq> S" by (auto simp: mem_iff)      
      from cond have "isOK(checkUrPClosedRM R (computeRuleMap R) ur P)" by auto
      hence "urClosed (set ur) R \<and> urPClosed R (set ur) (set P)" using checkUrPClosedRMSound by auto
      with cond id stri condR idr strir show ?thesis by auto
    qed
  qed
  show "set rR \<subseteq>ci set R \<and> (\<forall> s t \<sigma>. min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set rR) (shift s i) (shift t i) (shift \<sigma> i)))" (is "?goal P R rP rR")
  proof -
    let ?prop = "\<lambda> P R rP rR. monoRedPairUrProc S NS (set ur) P R (computeRuleMap R) rP rR \<longrightarrow> ?goal P R rP rR"
    from monoRedPairUrProc_ci have "\<forall> P R rP rR. ?prop P R rP rR"
      unfolding chain_identifying_proc_impl_def  .
    hence "\<forall> R rP rR. ?prop P R rP rR" ..
    hence "\<forall> rP rR. ?prop P R rP rR" ..
    hence "?prop P R rP rR" using ci_subsetI[of "set rR" "set R"] by force
    with mono show ?thesis by blast
  qed
qed
end

end
